"""
Command line utils functions.  All utility function
should be here.

Author Mus spyroot@gmail.com
"""
import argparse
import os
import warnings
from pathlib import Path
from typing import Optional


def _find_ids(obj, key, result):
    """Recursively search all keys
    :param obj:
    :param key:
    :return:
    """
    if obj is None:
        return

    for k in obj.keys():
        if isinstance(obj[k], dict):
            _find_ids(obj[k], key, result)
        if isinstance(obj[k], list):
            for e in obj[k]:
                _find_ids(e, key, result)
        else:
            if k == key:
                result.append(obj[k])


def find_ids(obj, key):
    """Recursively search key in object
    :param obj:
    :param key:
    :return:
    """
    result = []
    _find_ids(obj, key, result)
    return result


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def save_if_needed(filename: str, raw_data,
                   data_format: Optional[str] = "json",
                   indents: Optional[int] = 4,
                   save_dir: Optional[str] = None) -> None:
    """Save json raw data to a file.

    if filename already full path function will use that
    if caller provider both full path and save_dir arg
    save_dir will take precedence.

    Note function will add prefix a file either json or yaml.

    if caller provided save_dir optional arg function will
    create a dir.

    :param indents: default json ident relevant for a json only
    :param filename: a filename
    :param raw_data: a raw json string
    :param data_format: data format we save.  json, yaml, xml etc.
    :param save_dir: a directory where we need to save, if not exists will create.
    :return: Nothing
    :raise ValueError if invalid data format passed.
    """
    if filename is None or len(filename) == 0:
        return

    if raw_data is None:
        return

    target_dir = None
    if save_dir is not None and len(save_dir) > 0:
        save_dir_path = Path(save_dir).expanduser().resolve()
        if save_dir_path.exists() and save_dir_path.is_dir():
            target_dir = str(save_dir_path)
        if not save_dir_path.exists():
            os.makedirs(str(save_dir_path), exist_ok=True)
        if not save_dir_path.is_dir():
            raise ValueError("Save dir is file not a directory.")

    file_path = Path(filename).expanduser().resolve()
    if file_path.is_dir():
        warnings.warn("Can't save a result it to a file. "
                      "Please provide a filename, "
                      "not a path to a directory.")
        return

    if target_dir is not None:
        final_filename = target_dir + str(file_path.stem)
    else:
        final_filename = str(file_path)

    if data_format == "json":
        import json
        if '.json' in final_filename:
            final_filename = f"{final_filename}"
        else:
            final_filename = f"{final_filename}.json"
        with open(final_filename, 'w') as f:
            json.dump(raw_data, f, indent=indents)
    elif data_format == "yaml":
        import yaml
        if '.yaml' in final_filename:
            final_filename = f"{final_filename}"
        else:
            final_filename = f"{final_filename}.yaml"
        with open(final_filename, 'w') as file:
            yaml.dump(filename, file)
    else:
        ValueError("Unknown format")



"""iDRAC delete job from iDRAC action

Command provides the option to delete a
job from iDRAC.

Author Mus spyroot@gmail.com
"""
import argparse

from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, CommandResult


class JobWatch(IDracManager, scm_type=ApiRequestType.JobWatch,
               name='job_watch',
               metaclass=Singleton):
    """Command watch job progress.
    """

    def __init__(self, *args, **kwargs):
        super(JobWatch, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('--async', action='store_true', required=False, dest="do_async",
                                default=False, help="Will create a task and will not wait.")

        cmd_parser.add_argument('-j', '--job_id', required=True, dest="job_id", type=str,
                                default=None, help="Job id. Example JID_744718373591")

        cmd_parser.add_argument('-f', '--filename', required=False, type=str,
                                default="",
                                help="filename if we need to save a respond to a file.")

        help_text = "command watch a job"
        return cmd_parser, "job-watch", help_text

    def execute(self, job_id: str,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Executes delete job from iDRAC action

        python idrac_ctl.py del_job --job_id RID_744980379189

        :param job_id: iDRAC job_id JID_744718373591
        :param do_async: note async will subscribe to an event loop.
        :param verbose: enables verbose output
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"boot_source:{job_id} do_async:{do_async} filename:{filename}")
            print(f"the rest of args: {kwargs}")

        data = self.fetch_job(job_id)
        return CommandResult(data, None, None)



"""iDRAC storage cmd

Command provides the option to retrieve storage information for specific
controller disk from iDRAC.

Example
python idrac_ctl.py storage-get --storage_controller NonRAID.Slot.6-1

Expanded
python idrac_ctl.py storage-get --storage_controller NonRAID.Slot.6-1 -e

Filter by OEM
python idrac_ctl.py storage-get -c AHCI.Embedded.2-1 --filter Oem

Filter by Drives and Volumes
python idrac_ctl.py storage-get -c AHCI.Embedded.2-1 --filter Drives,Volumes

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional

from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton


class StorageView(IDracManager, scm_type=ApiRequestType.StorageViewQuery,
                  name='storage_get',
                  metaclass=Singleton):
    """iDRACs REST API fetch storage information.
    """

    def __init__(self, *args, **kwargs):
        super(StorageView, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser()
        cmd_parser.add_argument('-c', '--controller', required=False, type=str,
                                default="",
                                help="controller name.")

        cmd_parser.add_argument('--filter', required=False,
                                type=str, dest="data_filter",
                                default="",
                                help="Filter Controllers/Drives/Volumes. "
                                     "(Example filter both Driver "
                                     "and Volumes -c AHCI.Embedded.2-1 "
                                     "--filter Drives,Volumes")

        help_text = "command fetch the storage information"
        return cmd_parser, "storage-get", help_text

    @staticmethod
    def filter_by_keys(json_data, attr_filter: Optional[str]):
        """Filter attribute from json_data
        :param json_data:
        :param attr_filter:
        :return:
        """
        result = {}
        attr_filter = attr_filter.strip()
        if "," in attr_filter:
            attr_filters = attr_filter.split(",")
        else:
            attr_filters = [attr_filter]

        if len(attr_filters) > 0:
            json_data = dict((k, json_data[k])
                             for k in json_data for a in attr_filters
                             if a.lower() in k.lower())
            result.update(json_data)
        return result

    def execute(self,
                controller: Optional[str] = None,
                data_filter: Optional[str] = None,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_expanded: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Get storage controller details.
        :param data_filter:
        :param do_expanded:
        :param controller: if empty cmd will return list of controllers.
        :param verbose: enables verbose output
        :param do_async: will not block and return result as future.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:  json, xml etc.
        :return: named tuple CommandResult
        :raise: AuthenticationFailed, UnexpectedResponse
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        target_api = f"/redfish/v1/Systems/System.Embedded.1/Storage/{controller}"
        cmd_rest = self.base_query(target_api,
                                   filename=filename,
                                   do_async=do_async,
                                   do_expanded=do_expanded)
        data = cmd_rest.data
        if data_filter is not None and len(data_filter) > 0:
            data = self.filter_by_keys(cmd_rest.data, data_filter)

        return CommandResult(data, None, None)



"""iDRAC query jobs services

The DellJobService resource provides some actions to support
Job management functionality.  Query and discovery.

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional
from base import Singleton, ApiRequestType, IDracManager, CommandResult


class JobDellServices(IDracManager,
                      scm_type=ApiRequestType.JobDellServices,
                      name='job_service_query',
                      metaclass=Singleton):
    """A command query job_service_query.
    """

    def __init__(self, *args, **kwargs):
        super(JobDellServices, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser()
        help_text = "command query jobs services"
        return cmd_parser, "jobs-dell-service", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_expanded: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Executes query job services.
        python idrac_ctl.py query
        :param do_async: note async will subscribe to an event loop.
        :param do_expanded:  will do expand query
        :param filename: if filename indicate call will save a response to a file.
        :param verbose: enables verbose output
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        target_api = "/redfish/v1/Managers/iDRAC.Embedded.1/Oem/Dell/DellJobService"
        cmd_result = self.base_query(target_api,
                                     filename=filename,
                                     do_async=do_async,
                                     do_expanded=do_expanded)
        actions = self.discover_redfish_actions(self, cmd_result.data)
        return CommandResult(cmd_result, None, actions)



"""Main entry for idrac_ctl

The main interface consumed is iDRAC Manager class.
Each command registered dynamically and dispatch to respected execute method
by invoking request from IDRAC Manager.

Author Mus spyroot@gmail.com
"""

import argparse
import collections
import json
import os
import sys
import warnings
from typing import Optional, Dict

import requests
import urllib3
from pygments import highlight
try:
    from pygments.lexers.data import JsonLexer
except ImportError as ie:
    warnings.warn("Failed import json lexer from pygments.")

from pygments.formatters.terminal256 import Terminal256Formatter

from base.cmd_utils import save_if_needed
from base.shared import RedfishAction, RedfishActionEncoder
from base.cmd_exceptions import InvalidArgument, FailedDiscoverAction, UnsupportedAction
from base.idrac_manager import AuthenticationFailed, IDracManager, ResourceNotFound
from base.custom_argparser.customer_argdefault import CustomArgumentDefaultsHelpFormatter

try:
    from urllib3.exceptions import InsecureRequestWarning
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
except ImportError as ir:
    warnings.warn("Failed import urllib3")


def formatter(prog):
    argparse.HelpFormatter(prog, max_help_position=100, width=200)


class info:
    version = '0.8'
    description = 'iDRAC command line tools.'
    author = 'Mus'
    author_email = 'spyroot@gmail.com'


__version__ = info.version


def json_printer(json_data, cmd_args,
                 sort: Optional[bool] = True, indents: Optional[int] = 4,
                 colorized: Optional[bool] = True,
                 header: Optional[str] = None,
                 footer: Optional[str] = None) -> None:
    """Json stdout printer.
    :param cmd_args:
    :param header:
    :param footer:
    :param colorized:
    :param json_data:
    :param indents:
    :param sort:
    :return:
    """
    if cmd_args.no_stdout:
        return

    if isinstance(json_data, RedfishAction):
        json_raw = json_data.toJSON()
    elif isinstance(json_data, requests.models.Response):
        json_raw = json_data.json()
    elif isinstance(json_data, str):
        json_raw = json.dumps(json.loads(json_data),
                              sort_keys=sort, indent=indents)
    else:
        json_raw = json.dumps(json_data,
                              sort_keys=sort, indent=indents, cls=RedfishActionEncoder)

    if header is not None:
        print(header)

    if colorized:
        colorful = highlight(
                json_raw,
                lexer=JsonLexer(),
                formatter=Terminal256Formatter())
        print(colorful)
    else:
        print(json_raw)

    if footer is not None:
        print(footer)


def main(cmd_args, command_name_to_cmd: Dict) -> None:
    """Main handler
    """
    if args.insecure:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    redfish_api = IDracManager(idrac_ip=cmd_args.idrac_ip,
                               idrac_username=cmd_args.idrac_username,
                               idrac_password=cmd_args.idrac_password,
                               insecure=cmd_args.insecure)

    _ = redfish_api.check_api_version()

    if cmd_args.verbose:
        print("Verbose is set on")

    try:
        if cmd_args.subcommand in command_name_to_cmd:
            cmd = command_name_to_cmd[cmd_args.subcommand]
            arg_dict = dict((k, v) for k, v in vars(args).items() if k != "message_type")
            if cmd_args.verbose:
                print("# args dictionary:")
                json_printer(arg_dict, cmd_args)

            command_result = redfish_api.sync_invoke(cmd.type,
                                                     cmd.name,
                                                     **arg_dict)

            if cmd_args.json and command_result.data is not None:
                json_printer(command_result.data, cmd_args,
                             header="# respond data from the command:")

            # extra data for deep walks
            if command_result.extra is not None and cmd_args.no_extra is False:
                extra = command_result.extra
                if args.json:
                    json_printer(extra, cmd_args,
                                 header="#command extra data:")

                # save extra as separate files.
                if hasattr(cmd_args, 'do_save') and cmd_args.do_save:
                    for extra_k in extra.keys():
                        if args.verbose:
                            print(f"Saving {extra_k}.json")
                        save_if_needed(f"{extra_k}.json", extra[extra_k])

            # discovered rest action.
            if command_result.discovered is not None and cmd_args.no_action is False:
                if cmd_args.json:
                    if isinstance(command_result.discovered, dict):
                        for ak in command_result.discovered.keys():
                            if isinstance(command_result.discovered[ak], RedfishAction):
                                json_printer(json.dumps(command_result.discovered[ak].__dict__), cmd_args,
                                             header="# Redfish actions:")
                            else:
                                json_printer(json.dumps(command_result.discovered[ak]),
                                             cmd_args,
                                             header="# Redfish actions:")
                    else:
                        json_printer(command_result.discovered, cmd_args,
                                     header="# Redfish actions:")

    except ResourceNotFound as rnf:
        print("Error:", rnf)
    except InvalidArgument as ia:
        print("Error:", ia)
    except FailedDiscoverAction as fda:
        print("Error:", fda)
    except UnsupportedAction as ua:
        print("Error:", ua)


def create_cmd_tree(arg_parser) -> Dict:
    """Create command tree structure.
    :return:
    """
    redfish_api = IDracManager()
    command_name_to_cmd = {}
    commands_registry = redfish_api.get_registry()
    command_name = collections.namedtuple("Command", "type name")

    subparsers = arg_parser.add_subparsers(title='main command', metavar="main command",
                                           help='list of idrac_ctl commands',
                                           dest="subcommand",
                                           description='''Each action requires choosing
                                           a main command bios, boot, etc|n          
                                           Example: idrac_ctl.py bios\n''',
                                           required=True)

    for k in commands_registry:
        for sub_key in commands_registry[k]:
            cls = commands_registry[k][sub_key]
            if debug:
                print(f"Registering command {k} {sub_key}")
            if hasattr(cls, "register_subcommand"):
                cli_arg_parser, cmd_name, cmd_help = cls.register_subcommand(cls)

                # try:
                #     cli_arg_parser, cmd_name, cmd_help = cls.register_subcommand(cls, parent=arg_parser)
                # except TypeError:
                #     cli_arg_parser, cmd_name, cmd_help = cls.register_subcommand(cls)

                if debug:
                    print(f"Registering command name {cmd_name} {cmd_help}")
                subparsers.add_parser(cmd_name, parents=[cli_arg_parser], help=f"{str(cmd_help)}",
                                      formatter_class=CustomArgumentDefaultsHelpFormatter,
                                      )
                command_name_to_cmd[cmd_name] = command_name(k, sub_key)

    return command_name_to_cmd


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="idrac_ctl", add_help=True,
                                     description='''iDrac command line tools. |n
                                     It a standalone command line tool provide option to interact with  |n 
                                     Dell iDRAC via Redfish REST API. It supports both asynchronous and |n
                                     synchronous options to interact with iDRAC.|n
                                     Author Mus''',
                                     epilog="For more details example. Make sure to check."
                                            "https://github.com/spyroot/idrac_ctl",
                                     formatter_class=CustomArgumentDefaultsHelpFormatter)
    # global args
    parser.add_argument('--idrac_ip', required=False, type=str,
                        default=os.environ.get('IDRAC_IP', ''),
                        help="idrac ip address, by default "
                             "read from environment IDRAC_IP.")
    parser.add_argument('--idrac_username', required=False, type=str,
                        default=os.environ.get('IDRAC_USERNAME', 'root'),
                        help="idrac ip address, by default "
                             "read from environment IDRAC_USERNAME.")
    parser.add_argument('--idrac_password', required=False, type=str,
                        default=os.environ.get('IDRAC_PASSWORD', ''),
                        help="idrac ip address, by default "
                             "read from environment IDRAC_PASSWORD.")
    parser.add_argument('--insecure', action='store_true', required=False,
                        help="insecure ssl.")
    parser.add_argument('--debug', action='store_true', required=False,
                        help="enables debug.")
    parser.add_argument('--verbose', action='store_true', required=False, default=False,
                        help="enables verbose output.")
    parser.add_argument('--no_extra', action='store_true', required=False, default=False,
                        help="disables extra data stdout output.")
    parser.add_argument('--no_action', action='store_true', required=False, default=False,
                        help="disables rest action data stdout output.")
    parser.add_argument('--json', action='store_true', required=False, default=True,
                        help="by default we use json to output to console.")
    parser.add_argument('--no-stdout', '--no_stdout',  action='store_true', required=False, default=False,
                        help="by default we use stdout output.")
    parser.add_argument('-f', '--filename', required=False, type=str,
                        default="", help="Filename if we need save to a file.")

    debug = False
    cmd_dict = create_cmd_tree(parser)
    args = parser.parse_args()

    if args.idrac_ip is None or len(args.idrac_ip) == 0:
        print("Please indicate the idrac ip. "
              "--idrac_ip or set IDRAC_IP environment.")
        sys.exit(1)
    if args.idrac_username is None or len(args.idrac_username) == 0:
        print("Please indicate the idrac username. "
              "--idrac_username or set IDRAC_USERNAME environment.")
        sys.exit(1)
    if args.idrac_password is None or len(args.idrac_password) == 0:
        print("Please indicate the idrac password. "
              "--idrac_password or set IDRAC_PASSWORD environment.")
        sys.exit(1)

    try:
        main(args, cmd_dict)
    except AuthenticationFailed as af:
        print(f"Error: {af}")



"""iDRAC boot option query

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio

from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, CommandResult, save_if_needed


class BootOptionsQuery(IDracManager, scm_type=ApiRequestType.BootOptionQuery,
                       name='boot_options_query',
                       metaclass=Singleton):
    """
    Command enable boot option
    """

    def __init__(self, *args, **kwargs):
        super(BootOptionsQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('--async', default=False,  action='store_true',
                                required=False, dest="do_async",
                                help="Will create a task and will not wait.")

        cmd_parser.add_argument('-f', '--filename', required=False,
                                default="", type=str,
                                help="filename if we need to save a respond to a file.")

        help_text = "command fetch the boot options"
        return cmd_parser, "boot-option", help_text

    def execute(self,
                filename: Optional[str] = None,
                do_async: Optional[bool] = False,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Query information for particular boot source device from idrac.
        Example python idrac_ctl.py get_boot_source --dev "HardDisk.List.1-1"
        :param do_async: note async will subscribe to an event loop.
        :param verbose: enable verbose output.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"do_async:{do_async} filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        target_api = "/redfish/v1/Systems/System.Embedded.1/BootOptions"
        r = f"https://{self.idrac_ip}{target_api}{self.expanded()}"
        if not do_async:
            response = self.api_get_call(r, headers)
            self.default_error_handler(response)
        else:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

        data = response.json()
        save_if_needed(filename, data)
        return CommandResult(data, None, None)



"""iDRAC virtual disk query command

Command provides the option to retrieve virtual disk from iDRAC
and serialize back to caller as JSON, YAML, or XML. In addition, it automatically
registered to the command line ctl tool. Similarly to the rest command caller can save
to a file and consume asynchronously or synchronously.

- Each command return a result and list of REST Actions.
- Each command loaded based __init__ hence anyone can extend and add custom command.

"RAID.Mezzanine.1-1"
"RAID.SL.3-1"
"Disk.Bay.2:Enclosure.Internal.0-1:RAID.Mezzanine.1-1"

Author Mus spyroot@gmail.com
"""
import argparse
import warnings
from abc import abstractmethod
from typing import Optional

from base import CommandResult, save_if_needed, InvalidArgument
from base import IDracManager, ApiRequestType, Singleton
from base.idrac_manager import ResourceNotFound


class VirtualDiskQuery(IDracManager, scm_type=ApiRequestType.VirtualDiskQuery,
                       name='virtual_disk_query',
                       metaclass=Singleton):
    """iDRACs REST API Virtual Disk Query Command, fetch virtual disk, caller can save
    result to a file or output stdout or pass downstream to jq etc. tools.
    """

    def __init__(self, *args, **kwargs):
        super(VirtualDiskQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)
        cmd_arg.add_argument('--device_id', required=False, type=str,
                             default="", help="storage device id. "
                                              "Example NonRAID.Slot.6-1.")

        cmd_arg.add_argument('-f', '--filename', required=False, type=str,
                             default="",
                             help="filename if we need to save a respond to a file.")

        help_text = "fetch the virtual disk data"
        return cmd_arg, "volumes", help_text

    def execute(self,
                filename: Optional[str] = None,
                device_id: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Queries volumes from given controlled from iDRAC.
        :param device_id: a storage controller id, NonRAID.Slot.6-1
        :param verbose: enables verbose output
        :param do_deep: do deep recursive fetch
        :param do_async: will not block and return result as future.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:  json, xml etc.
        :return: named tuple CommandResult
        :raise: AuthenticationFailed, UnexpectedResponse
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        storage_id = self.sync_invoke(ApiRequestType.StorageListQuery,
                                      "storage_list")
        storage_members = storage_id.data['Members']
        storage_ids = [k['@odata.id'].split("/")[-1]
                       for k in storage_members if '@odata.id' in k]

        if device_id not in storage_ids:
            raise InvalidArgument(f"Storage device_id {device_id} "
                                  "not found, available {storage_ids}")

        r = f"https://{self.idrac_ip}/redfish/v1/Systems/" \
            f"System.Embedded.1/Storage/{device_id}/Volumes"
        #
        response = self.api_get_call(r, headers)
        self.default_error_handler(response)
        data = response.json()

        vd_list = []
        if not data['Members']:
            return CommandResult(None, None, None)
        else:
            vd_list = [i['@odata.id'].split("/")[-1] for i in data['Members']]

        for vol_id in vd_list:
            r = f"https://{self.idrac_ip}/redfish/v1/Systems" \
                f"/System.Embedded.1/Storage/Volumes/{vol_id}"
            try:
                response = self.api_get_call(r, headers)
                self.default_error_handler(response)
            except ResourceNotFound as exp:
                warnings.warn(str(exp))
                continue
                pass
            resp_data = response.json()
            vd_list.append(resp_data)

        save_if_needed(filename, data)
        return CommandResult(vd_list, None, None)



"""iDRAC Redfish API with Dell OEM extension
to get network ISO attach status.

python idrac_ctl.py oem-attach --ip_addr 10.241.7.99 \
--share_name sambashare --remote_image ubuntu-22.04.1-desktop-amd64.iso

❯ python idrac_ctl.py oem-attach-status
# respond data from the command:
{
    "DriversAttachStatus": "NotAttached",
    "ISOAttachStatus": "Attached"
}

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional
from base import Singleton, ApiRequestType, IDracManager, CommandResult


class DellOemAttach(IDracManager, scm_type=ApiRequestType.OemAttach,
                    name='delloem_attach',
                    metaclass=Singleton):
    """A command uses dell oem to attach ISO
    """

    def __init__(self, *args, **kwargs):
        super(DellOemAttach, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser(is_remote_share=True)
        help_text = "command attach network iso "
        return cmd_parser, "oem-attach", help_text

    def execute(self,
                ip_addr: Optional[str] = None,
                share_type: Optional[str] = None,
                share_name: Optional[str] = None,
                remote_image: Optional[str] = None,
                remote_username: Optional[str] = None,
                remote_password: Optional[str] = None,
                remote_workgroup: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Executes dell oem ConnectNetworkISOImage

        :param ip_addr: ip address of NFS or CIFS
        :param share_type: NFS|CIFS
        :param share_name: share name
        :param remote_image:  path to image
        :param remote_username: remote username if required for NFS or CIFS
        :param remote_password: remote password if required for NFS or CIFS
        :param remote_workgroup:
        :param do_async: note async will subscribe to an event loop.
        :param verbose: enables verbose output
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        cmd_result = self.sync_invoke(ApiRequestType.DellOemActions, "dell_oem_actions")
        redfish_action = cmd_result.discovered['ConnectNetworkISOImage']
        target_api = redfish_action.target
        # args = redfish_action.args

        payload = {
            'IPAddress': ip_addr,
            'ShareType': share_type,
            'ShareName': share_name,
            'ImageName': remote_image,
            'UserName': remote_username,
            'Password': remote_password,
            'Workgroup': remote_workgroup,
        }

        for key, value in dict(payload).items():
            if value is None:
                del payload[key]

        api_result = self.base_post(target_api, payload=payload,
                                    do_async=do_async, expected_status=202)
        result = {}
        if api_result is not None and api_result.extra is not None:
            data = api_result.extra.json()
            self.default_json_printer(data)

        return CommandResult(result, None, None)



"""iDRAC import system config from json file.

Command provides the option to import configuration.

python idrac_ctl.py system-export --filename system.json
python idrac_ctl.py system-import --config system.json

Author Mus spyroot@gmail.com
"""
import json
from abc import abstractmethod
from pathlib import Path
from typing import Optional

from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton
from base.cmd_exceptions import InvalidArgument
from base.idrac_manager import PostRequestFailed


class ImportOneTimeBoot(IDracManager,
                        scm_type=ApiRequestType.ImportOneTimeBoot,
                        name='import_sysconfig',
                        metaclass=Singleton):
    """
    Command implementation import system configuration.
    """
    def __init__(self, *args, **kwargs):
        super(ImportOneTimeBoot, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command
        :param cls:
        :return:
        """
        cmd_arg = cls.base_parser(is_reboot=True, is_expanded=False)

        cmd_arg.add_argument('--config', required=True, dest="config", type=str,
                             default=None, help="Import system config")

        cmd_arg.add_argument('--shutdown_type',
                             required=False, type=str, default="Graceful",
                             help="Graceful, Forced, NoReboot.")

        cmd_arg.add_argument('--host_power_state',
                             required=False, type=str, default="On",
                             help="Graceful, Forced, NoReboot.")

        cmd_arg.add_argument('--time_to_wait',
                             required=False, type=int, default=300,
                             help="The time to wait for the host to shut down. "
                                  "Default and minimum value is 300 seconds. "
                                  "Maximum value is 3600 seconds..")

        help_text = "command import system configuration"
        return cmd_arg, "system-import", help_text

    def execute(self,
                config: str,
                shutdown_type: Optional[str] = "Graceful",
                host_power_state: Optional[str] = "Off",
                time_to_wait: Optional[str] = "",
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_reboot: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Alternative method to set boot from cdrom.

        :param do_reboot:
        :param host_power_state: On, Off
        :param config: path to a config file.
        :param shutdown_type:  Graceful, Forced, NoReboot.
        :param time_to_wait:
        :param do_async: will schedule asyncio task.
        :param filename: if filename indicate call will save respond to a file.
        :param verbose: verbose output.
        :param data_type: a data serialized back.
        :return: in data type json will return json
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"shutdown_type:{shutdown_type} "
                  f"host_power_state:{host_power_state} "
                  f"do_async:{do_async} "
                  f"filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        shutdown_types = ['Graceful', 'Forced', 'NoReboot']
        host_power_states = ['On', 'Off']

        if shutdown_type.title() not in shutdown_type:
            raise InvalidArgument(f"Invalid shutdown type "
                                  f"{shutdown_type} "
                                  f"supported {shutdown_types}")

        if host_power_state.title() not in host_power_states:
            raise InvalidArgument(f"Invalid power state type "
                                  f"{host_power_states} "
                                  f"supported {host_power_states}")

        path_config = Path(config).expanduser().resolve()
        if not path_config.is_file():
            raise InvalidArgument(f"Invalid path to a config file.")

        buf = "<SystemConfiguration>""<Component FQDD=\"iDRAC.Embedded.1\">" \
              "<Attribute Name=\"ServerBoot.1#BootOnce\">Enabled</Attribute>" \
              "<Attribute Name=\"ServerBoot.1#FirstBootDevice\">VCD-DVD</Attribute>" \
              "</Component></SystemConfiguration>"

        payload = {"ShutdownType": shutdown_type.title(),
                   "HostPowerState": host_power_state.title(),
                   "ImportBuffer": buf,
                   "ShareParameters": {"Target": "ALL"}
                   }

        r = f"https://{self.idrac_ip}/redfish/v1/Managers/iDRAC.Embedded.1/" \
            f"Actions/Oem/EID_674_Manager.ImportSystemConfiguration"

        data = {}
        try:
            response = self.api_post_call(r, json.dumps(payload), headers)
            ok = self.default_post_success(self, response, expected=202)

            if ok:
                job_id = self.job_id_from_header(response)
                if job_id is None:
                    self.job_id_from_respond(response)
                if job_id is not None:
                    if not do_async:
                        data = self.fetch_job(job_id)
                    else:
                        data = {"job_id": job_id}

            if do_reboot:
                reboot_result = self.reboot()
                data.update(reboot_result)
        except PostRequestFailed as prf:
            print(prf)

        return CommandResult(data, None, None)



"""iDRAC attribute command

Command provides the option to retrieve the iDRAC attribute and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command caller can save
to a file and consume asynchronously or synchronously.

python idrac_ctl.py --json attribute --filter ServerPwrMon.1.PeakCurrentTime

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
from abc import abstractmethod
from typing import Optional

from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton
from base.cmd_utils import save_if_needed, find_ids


class AttributesQuery(IDracManager, scm_type=ApiRequestType.AttributesQuery,
                      name='attribute_inventory',
                      metaclass=Singleton):
    """iDRAC Attribute Query Command, fetch attribute data, caller can save to a file
    or output to a file or pass downstream.
    """

    def __init__(self, *args, **kwargs):
        super(AttributesQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)

        cmd_arg.add_argument('--async', action='store_true', required=False, dest="do_async",
                             default=False, help="Will create a task and will not wait.")

        cmd_arg.add_argument('--deep', action='store_true', required=False, dest="do_deep",
                             default=False, help="deep walk. will make a separate "
                                                 "REST call for each rest api.")

        cmd_arg.add_argument('--attr_only', action='store_true', default=False,
                             required=False, dest='attr_only',
                             help="will show only attributes.")

        cmd_arg.add_argument('--filter', required=False, type=str, dest="attr_filter",
                             help="Filter on bios attribute information. "
                                  "Example --filter ProcCStates , "
                                  "will filter and and show C-State.")

        cmd_arg.add_argument('-f', '--filename', required=False, type=str,
                             default="",
                             help="filename if we need to save a respond to a file.")

        help_text = "command fetch the attribute view"
        return cmd_arg, "attr", help_text

    def execute(self, filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                attr_only: Optional[bool] = False,
                attr_filter: Optional[str] = "",
                **kwargs) -> CommandResult:
        """Queries attributes from iDRAC.
        :param attr_filter: filter on specific attribute.
        :param attr_only:
        :param do_async:
        :param verbose:
        :param do_deep:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:
        :return:
        :raise: AuthenticationFailed, UnexpectedResponse
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        t = "/redfish/v1/Managers/System.Embedded.1/Attributes"
        r = f"https://{self.idrac_ip}{t}"
        if not do_async:
            response = self.api_get_call(r, headers)
            self.default_error_handler(response)
        else:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

        data = response.json()

        # filter
        data = self.filter_attribute(self, data, attr_filter)
        extra_actions = find_ids(data, "@odata.id")
        extra_data = None
        if do_deep:
            extra_data = [self.api_get_call(f"https://{self.idrac_ip}{a}", headers).json()
                          for a in extra_actions]

        save_if_needed(filename, data)
        return CommandResult(data, None, extra_data)



"""iDRAC boot source command

Command provides the option to retrieve boot source from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command
caller can save to a file and consume asynchronously or synchronously.

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
from abc import abstractmethod
from typing import Optional
from base import Singleton, ApiRequestType
from base import IDracManager, CommandResult
from base.cmd_utils import save_if_needed, find_ids


class BootQuery(IDracManager, scm_type=ApiRequestType.BootQuery,
                name='boot_query',
                metaclass=Singleton):
    """
    Command return boot source
    """
    def __init__(self, *args, **kwargs):
        super(BootQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)
        cmd_arg.add_argument('--async', action='store_true',
                             required=False, dest="do_async",
                             default=False,
                             help="Will use asyncio.")

        cmd_arg.add_argument('-f', '--filename', required=False, type=str,
                             default="",
                             help="filename if we need to save a respond to a file.")

        cmd_arg.add_argument('--deep', action='store_true', required=False, dest="do_deep",
                             default=False, help="deep walk. will make a separate "
                                                 "rest call for each discovered api.")

        help_text = "command fetch the boot source"
        return cmd_arg, "boot", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Query boot source from idrac
        :param do_async: will use asyncio
        :param verbose:
        :param do_deep:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r = f"https://{self.idrac_ip}/redfish/v1/Systems" \
            f"/System.Embedded.1/BootSources"

        if not do_async:
            response = self.api_get_call(r, headers)
            self.default_error_handler(response)
        else:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

        data = response.json()
        save_if_needed(filename, data)

        # extra data
        extra_actions = find_ids(data, "@odata.id")
        extra_data = None
        if do_deep:
            extra_data = [self.api_get_call(f"https://{self.idrac_ip}{a}", headers).json()
                          for a in extra_actions]

        return CommandResult(data, None, extra_data)



"""iDRAC enable boot options.

This cmd return Dell Boot Sources Configuration and the related
resources for particular device

Example.
python idrac_ctl.py boot-source-get --dev NIC.Slot.8-1


Author Mus spyroot@gmail.com
"""
import argparse
import asyncio

from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, CommandResult, save_if_needed


class EnableBootOptions(IDracManager,
                        scm_type=ApiRequestType.EnableBootOptions,
                        name='boot_source_query',
                        metaclass=Singleton):
    """
    Command fetch boot option for particular device.
    """

    def __init__(self, *args, **kwargs):
        super(EnableBootOptions, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False,
                                             description="command fetch the boot source for device/devices")
        # idrac_ctl.py boot-source-get --dev NIC.Slot.8-1
        cmd_parser.add_argument('--async', action='store_true', required=False,
                                dest="do_async", default=False,
                                help="will use async task and will not wait")

        cmd_parser.add_argument('--dev', required=False, dest="boot_source",
                                type=str, default=None,  metavar="DEVICE",
                                help="fetch verbose information for a device. Example --dev NIC.Slot.8-1")

        cmd_parser.add_argument('-f', '--filename', required=False,
                                type=str, default="",
                                help="filename if we need to save a respond to a file.")

        help_text = "command fetch the boot source for device/devices"
        return cmd_parser, "boot-source-get", help_text

    def execute(self,
                boot_source: Optional[str] = None,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Query information for particular boot source device from idrac.
        Example python idrac_ctl.py get_boot_source --dev "HardDisk.List.1-1"
        :param boot_source: a device HardDisk.List.1-1
        :param do_async: note async will subscribe to an event loop.
        :param verbose:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"boot_source:{boot_source} do_async:{do_async} filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        result = self.sync_invoke(ApiRequestType.BootOptions, "boot_sources_query")
        boot_data_sources = {}
        for full_dev_path in result.data:
            r = f"https://{self.idrac_ip}{full_dev_path}?$expand=*($levels=1)"
            if not do_async:
                response = self.api_get_call(r, headers)
                self.default_error_handler(response)
            else:
                loop = asyncio.get_event_loop()
                response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

            dev_data = response.json()
            devs = full_dev_path.split("/")
            if len(devs) > 0:
                dev = devs[-1]
                # add all by default
                if boot_source is None:
                    boot_data_sources[dev] = dev_data
                elif boot_source is not None and boot_source.lower() in dev.lower():
                    boot_data_sources[dev] = dev_data
                    break

        save_if_needed(filename, boot_data_sources)
        return CommandResult(boot_data_sources, None, None)



"""iDRAC storage list view

Command provides the option to retrieve list of storage controllers.

Example expanded
python idrac_ctl.py storage-list -e

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional

from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton


class StorageListView(IDracManager, scm_type=ApiRequestType.StorageListQuery,
                      name='storage_list',
                      metaclass=Singleton):
    """iDRACs REST API fetch storage list information.
    """

    def __init__(self, *args, **kwargs):
        super(StorageListView, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser()
        help_text = "command fetch the storage devices"
        return cmd_parser, "storage-list", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_expanded: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Queries for storage controller list.
        :param do_expanded:
        :param do_async: will not block and return result as future.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:  json, xml etc.
        :param verbose: enables verbose output
        :return: named tuple CommandResult
        :raise: AuthenticationFailed, UnexpectedResponse
        """
        target_api = "/redfish/v1/Systems/System.Embedded.1/Storage"
        return self.base_query(target_api,
                               filename=filename,
                               do_async=do_async,
                               do_expanded=do_expanded)



"""iDRAC boot options.

Command provides the option to retrieve boot source from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command
caller can save to a file and consume asynchronously or synchronously.

Doc
https://www.dell.com/support/manuals/en-us/idrac9-lifecycle-controller-v3.3-series/
idrac9_3.36_redfishapiguide/dellbootsources?guid=guid-4803ff0e-76ad-42c5-a971-820123cd0b83&lang=en-us

python idrac_ctl.py --json boot_source

cmd return list of boot devices.

    "/redfish/v1/Systems/System.Embedded.1/BootOptions/HardDisk.List.1-1",
    "/redfish/v1/Systems/System.Embedded.1/BootOptions/NIC.Integrated.1-1-1",
    "/redfish/v1/Systems/System.Embedded.1/BootOptions/NIC.Slot.8-1",
    "/redfish/v1/Systems/System.Embedded.1/BootOptions/NIC.Slot.8-1",
    "/redfish/v1/Systems/System.Embedded.1/BootOptions/NIC.Slot.8-1",
    "/redfish/v1/Systems/System.Embedded.1/BootOptions/NIC.Slot.8-1"


Author Mus spyroot@gmail.com
"""
import argparse
from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, CommandResult


class BootOptionsList(IDracManager, scm_type=ApiRequestType.BootOptions,
                      name='boot_sources_query',
                      metaclass=Singleton):
    """
    Command enable boot option
    """

    def __init__(self, *args, **kwargs):
        super(BootOptionsList, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('--async', action='store_true', required=False, dest="do_async",
                                default=False, help="Will create a task and will not wait.")

        cmd_parser.add_argument('-f', '--filename', required=False, type=str,
                                default="",
                                help="filename if we need to save a respond to a file.")

        help_text = "command fetch the boot source list"
        return cmd_parser, "boot-source-list", help_text

    def execute(self, filename: [str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False, **kwargs) -> CommandResult:
        """List boot source from idrac
        :param do_async:
        :param verbose:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r = f"https://{self.idrac_ip}/redfish/v1/Systems" \
            f"/System.Embedded.1/BootOptions?$expand=*($levels=1)"

        response = self.api_get_call(r, headers)
        data = response.json()
        extra = data
        self.default_error_handler(response)

        if 'Members' in data:
            data = data['Members']
            data = [d['@odata.id'] for d in data]

        return CommandResult(data, None, extra)



"""iDRAC export system config command.

Command provides the option to retrieve firmware setting from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command caller can save
to a file and consume asynchronously or synchronously.

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
import json
from abc import abstractmethod
from typing import Optional

from base import IDracManager, ApiRequestType, Singleton
from base.cmd_utils import save_if_needed
from base.idrac_manager import UnexpectedResponse, CommandResult


class ExportSystemConfig(IDracManager,
                         scm_type=ApiRequestType.SystemConfigQuery,
                         name='sysconfig_query',
                         metaclass=Singleton):
    """
    Command exports system configuration..
    """

    def __init__(self, *args, **kwargs):
        super(ExportSystemConfig, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command arguments.
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('--async', action='store_true',
                                required=False, dest="do_async",
                                default=False,
                                help="Will create a task and will not wait.")

        cmd_parser.add_argument('-f', '--filename', required=False, type=str,
                                default="",
                                help="filename if we need to save a respond to a file.")

        cmd_parser.add_argument('--export_use', action='store_true',
                                required=False,  default=False,
                                help="Will create a task and will not wait.")

        help_text = "command exports system configuration"
        return cmd_parser, "system-export", help_text

    def execute(self,
                data_type: Optional[str] = "json",
                filename: Optional[str] = None,
                export_format: Optional[str] = "json",
                export_use: Optional[str] = "Default",
                include_in_export: Optional[str] = "Default",
                target: Optional[str] = "ALL",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs
                ) -> CommandResult:
        """Exports system configuration from idrac

        Default, Clone, Replace
        Default, IncludeReadOnly, IncludePasswordHashValues
        ShareParameters

        Share parameters and values
        - IP address of the network share
        - Name of network share
        - File name for the SCP
        - CIFS, NFS, HTTP, HTTPS
        - Username to log on to the share — for CIFS share only.
        - Password to log on to the share — for CIFS share only.
        - Workgroup name to log on to the share
        - Can be the component name or an FQDN. The default value is ALL.

        :param do_async:
        :param data_type:
        :param verbose:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param include_in_export:
        :param target:
        :param export_format:
        :param export_use:
        :return:
        """

        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"do_async:{do_async} export_format:{export_format}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        headers.update(self.content_type)
        payload = {"ExportFormat": export_format.upper(),
                   "ShareParameters": {"Target": target},
                   "IncludeInExport": include_in_export}

        if "Clone" in export_use or "Replace" in export_use:
            payload["ExportUse"] = export_use

        r = f"https://{self.idrac_ip}/redfish/v1/Managers/iDRAC.Embedded.1/" \
            f"Actions/Oem/EID_674_Manager.ExportSystemConfiguration"

        json_pd = json.dumps(payload)

        if not do_async:
            response = self.api_post_call(r, json_pd, headers)
            self.default_post_success(self, response)
        else:
            loop = asyncio.get_event_loop()
            ok, response = loop.run_until_complete(self.async_post_until_complete(r, json_pd, headers))

        resp_hdr = response.headers
        if 'Location' not in resp_hdr:
            raise UnexpectedResponse("rest api failed.")

        location = response.headers['Location']
        job_id = location.split("/")[-1]
        if not do_async:
            data = self.fetch_job(job_id)
            # we save only if we need to.
            if verbose:
                print(f"Saving to a file {filename}")
            save_if_needed(filename, data)
        else:
            data = {"job_id": job_id}

        return CommandResult(data, None, None)



"""iDRAC firmware inventory command

Command provides the option to retrieve firmware setting from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command caller can save
to a file and consume asynchronously or synchronously.

Author Mus spyroot@gmail.com
"""
import argparse
from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, save_if_needed, CommandResult


class FirmwareInventoryQuery(IDracManager, scm_type=ApiRequestType.FirmwareInventoryQuery,
                             name='firmware_inv_query',
                             metaclass=Singleton):
    """
    Command implementation to get current firmware version from iDRAC.
    """

    def __init__(self, *args, **kwargs):
        super(FirmwareInventoryQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command.
        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)
        cmd_arg.add_argument('--async', action='store_true', required=False, dest="do_async",
                             default=False, help="Will do async request.")

        cmd_arg.add_argument('--deep', action='store_true', required=False, dest="do_deep",
                             default=False, help="deep view to each pci.")

        cmd_arg.add_argument('-f', '--filename', required=False, type=str,
                             default="",
                             help="filename if we need to save a respond to a file.")

        help_text = "fetch the firmware inventory view"
        return cmd_arg, "firmware_inventory", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False, **kwargs) -> CommandResult:
        """Query firmware from idrac
        :param do_deep: will return verbose output for each pci device.
        :param do_async: will schedule asyncio task.
        :param verbose: verbose output.
        :param filename: if filename indicate call will save respond to a file.
        :param data_type: a data serialized back.
        :return: in data type json will return json
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)
        r = f"https://{self.idrac_ip}/redfish/v1/UpdateService/" \
            f"FirmwareInventory?$expand=*($levels=1)"
        response = self.api_get_call(r, headers)
        data = response.json()
        self.default_error_handler(response)
        save_if_needed(filename, data)
        return CommandResult(data, None, None)



"""iDRAC firmware command

Command provides the option to retrieve firmware setting from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command caller can save
to a file and consume asynchronously or synchronously.

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
from abc import abstractmethod
from typing import Optional

from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton
from base.cmd_utils import save_if_needed


class FirmwareQuery(IDracManager, scm_type=ApiRequestType.FirmwareQuery,
                    name='firmware_query',
                    metaclass=Singleton):
    """
    Command implementation to get current firmware version from iDRAC.
    """

    def __init__(self, *args, **kwargs):
        super(FirmwareQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """

        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)
        cmd_arg.add_argument('--async', action='store_true',
                             required=False, dest="do_async",
                             default=False, help="will do async request.")

        cmd_arg.add_argument('--deep', action='store_true',
                             required=False, dest="do_deep",
                             default=False, help="deep view to each pci.")

        cmd_arg.add_argument('-f', '--filename',
                             required=False, type=str, default="",
                             help="filename, if we need save to a file.")

        help_text = "fetch the firmware view"
        return cmd_arg, "firmware", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False, **kwargs) -> CommandResult:
        """Query firmware from idrac
        :param do_deep: will return verbose output for each pci device.
        :param do_async: will schedule asyncio task.
        :param verbose: verbose output.
        :param filename: if filename indicate call will save respond to a file.
        :param data_type: a data serialized back.
        :return: in data type json will return json
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"do_deep:{do_deep} do_async:{do_async} filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        if do_deep:
            r = f"https://{self.idrac_ip}/redfish/v1/UpdateService/" \
                f"FirmwareInventory?$expand=*($levels=1)"
        else:
            r = f"https://{self.idrac_ip}/redfish/v1/UpdateService/FirmwareInventory"

        if not do_async:
            response = self.api_get_call(r, headers)
            self.default_error_handler(response)
        else:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

        self.default_error_handler(response)
        data = response.json()

        save_if_needed(filename, data)
        return CommandResult(data, None, None)



"""iDRAC import system config from json file.

Command provides the option to import configuration.

python idrac_ctl.py system-export --filename system.json
python idrac_ctl.py system-import --config system.json

Author Mus spyroot@gmail.com
"""
import json
from abc import abstractmethod
from pathlib import Path
from typing import Optional

from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton
from base.cmd_exceptions import InvalidArgument
from base.idrac_manager import PostRequestFailed


class ImportSystemConfig(IDracManager, scm_type=ApiRequestType.ImportSystem,
                         name='import_sysconfig',
                         metaclass=Singleton):
    """
    Command implementation import system configuration.
    """

    def __init__(self, *args, **kwargs):
        super(ImportSystemConfig, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register system import cmd
        :param cls:
        :return:
        """
        cmd_arg = cls.base_parser(is_reboot=True, is_expanded=False)

        cmd_arg.add_argument('--config', required=True, dest="config", type=str,
                             default=None, help="Import system config")

        cmd_arg.add_argument('--shutdown_type',
                             required=False, type=str, default="Graceful",
                             help="Graceful, Forced, NoReboot.")

        cmd_arg.add_argument('--host_power_state',
                             required=False, type=str, default="On",
                             help="Graceful, Forced, NoReboot.")

        cmd_arg.add_argument('--time_to_wait',
                             required=False, type=int, default=300,
                             help="The time to wait for the host to shut down. "
                                  "Default and minimum value is 300 seconds. "
                                  "Maximum value is 3600 seconds..")

        help_text = "command import system configuration"
        return cmd_arg, "system-import", help_text

    def execute(self,
                config: str,
                shutdown_type: Optional[str] = "Graceful",
                host_power_state: Optional[str] = "Off",
                time_to_wait: Optional[str] = "",
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_reboot: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Import system config

        ImportBuffer
        Buffer content to perform import. Required only for LOCAL
        and not required for CIFS, NFS, HTTP, or HTTPS.

        ShutdownType ShutdownType Graceful, Forced, NoReboot

        HostPowerState  On, Off

        TimeToWait The time to wait for the host to shut down. Default and
        minimum value is 300 seconds. Maximum value is 3600
        seconds.

        Share parameters and values
        IPAddress IP address of the network share
        ShareName Name of network share
        File name for the SCP
        LOCAL, CIFS, NFS, HTTP, HTTPS


        ExportFormat
        ExportUse
        IncludeInExport
        ShareParameters

        :param do_reboot:
        :param host_power_state: On, Off
        :param config: path to a config file.
        :param shutdown_type:  Graceful, Forced, NoReboot.
        :param time_to_wait:
        :param do_async: will schedule asyncio task.
        :param filename: if filename indicate call will save respond to a file.
        :param verbose: verbose output.
        :param data_type: a data serialized back.
        :return: in data type json will return json
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"shutdown_type:{shutdown_type} "
                  f"host_power_state:{host_power_state} "
                  f"do_async:{do_async} "
                  f"filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        shutdown_types = ['Graceful', 'Forced', 'NoReboot']
        host_power_states = ['On', 'Off']

        if shutdown_type.title() not in shutdown_type:
            raise InvalidArgument(f"Invalid shutdown type "
                                  f"{shutdown_type} supported {shutdown_types}")

        if host_power_state.title() not in host_power_states:
            raise InvalidArgument(f"Invalid power state type "
                                  f"{host_power_states} supported {host_power_states}")

        path_config = Path(config).expanduser().resolve()
        if not path_config.is_file():
            raise InvalidArgument(f"Invalid path to a config file.")

        with open(str(path_config), "r") as f:
            buf = f.read()
            buf = buf.replace('\n', "")
            buf = buf.replace("   ", "")

        payload = {"ShutdownType": shutdown_type.title(),
                   "HostPowerState": host_power_state.title(),
                   "ImportBuffer": buf,
                   "ShareParameters": {"Target": "ALL"}
                   }

        r = f"https://{self.idrac_ip}/redfish/v1/Managers/iDRAC.Embedded.1/" \
            f"Actions/Oem/EID_674_Manager.ImportSystemConfiguration"

        data = {}

        try:
            response = self.api_post_call(r, json.dumps(payload), headers)
            ok = self.default_post_success(self, response, expected=202)

            if ok:
                job_id = self.job_id_from_header(response)
                if job_id is not None:
                    if not do_async:
                        data = self.fetch_job(job_id)
                    else:
                        data = {"job_id": job_id}

            if do_reboot:
                reboot_result = self.reboot()
                data.update(reboot_result)
        except PostRequestFailed as prf:
            print(prf)

        return CommandResult(data, None, None)



class InvalidArgument(Exception):
    pass


class FailedDiscoverAction(Exception):
    """Must raise if failed discover action.
    """
    pass


class NoPendingValues(Exception):
    """Must raise if failed discover action.
    """
    pass


class UnsupportedAction(Exception):
    pass



"""iDRAC reset a chassis power state.

python idrac_ctl.py chassis-reset --reset_type ForceOff

Author Mus spyroot@gmail.com
"""
import warnings
from abc import abstractmethod
from typing import Optional

from base import IDracManager, ApiRequestType, Singleton, UnexpectedResponse
from base import CommandResult


class ChangeBootOrder(IDracManager,
                      scm_type=ApiRequestType.ChangeBootOrder,
                      name='change_boot_order',
                      metaclass=Singleton):
    """Command change boot order.
    """

    def __init__(self, *args, **kwargs):
        super(ChangeBootOrder, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser(is_file_save=False, is_expanded=False)
        cmd_parser.add_argument('--boot_order',
                                required=False, dest='reset_type',
                                default="", type=str,
                                help="Coma seperated list. Example HardDisk.List.1-1,NIC.Integrated.1-1-1")

        help_text = "command change boot order"
        return cmd_parser, "change-boot-order", help_text

    def execute(self,
                boot_order: str,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_async: Optional[bool] = False,
                **kwargs
                ) -> CommandResult:
        """
        "Boot": {
        "BootOptions": {
            "@odata.id": "/redfish/v1/Systems/System.Embedded.1/BootOptions"
        },
        "BootOrder": [
            "HardDisk.List.1-1",
            "NIC.Integrated.1-1-1",
            "NIC.Slot.8-1",
            "NIC.Slot.8-1",
            "NIC.Slot.8-1",
            "NIC.Slot.8-1",
            "Optical.iDRACVirtual.1-1"
        ],
        "BootOrder@odata.count": 7,
        "BootSourceOverrideEnabled": "Disabled",
        "BootSourceOverrideMode": "Legacy",
        "BootSourceOverrideTarget": "None",
        "BootSourceOverrideTarget@Redfish.AllowableValues": [
            "None",
            "Pxe",
            "Floppy",
            "Cd",
            "Hdd",
            "BiosSetup",
            "Utilities",
            "UefiTarget",
            "SDCard",
            "UefiHttp"
        ],
        "Certificates": {
            "@odata.id": "/redfish/v1/Systems/System.Embedded.1/Boot/Certificates"
        },
        "UefiTargetBootSourceOverride": null

        :param boot_order:
        :param filename:
        :param do_async:
        :param data_type:
        :param kwargs:
        :return: return cmd result
        :raise FailedDiscoverAction
        """
        target_api = "/redfish/v1/Systems/System.Embedded.1/Bios"
        cmd_rest = self.base_query(target_api,
                                   filename=filename,
                                   do_async=do_async,
                                   do_expanded=False)

        self.default_json_printer(cmd_rest.data)
        current_boot_mode = cmd_rest.data['Attributes']['SetBootOrderEn']
        print("Current boot mode", current_boot_mode)

        if boot_order is not None:
            boot_order = boot_order.strip().split(",")

        payload = {
            "Boot": {
                "BootOrder": boot_order
            }
        }

        target_patch_api = "/redfish/v1/Systems/System.Embedded.1"
        api_result = self.base_patch(target_patch_api, payload=payload,
                                     do_async=do_async, expected_status=202)
        result = api_result.data
        if api_result is not None and api_result.extra is not None:
            resp = api_result.extra
            data = api_result.extra.json()
            self.default_json_printer(data)
            try:
                job_id = self.job_id_from_header(resp)
                if job_id is not None:
                    data = self.fetch_job(job_id)
                    result = data
            except UnexpectedResponse as ur:
                warnings.warn(str(ur))
                pass

        return CommandResult(result, None, None)



"""iDRAC Redfish API with Dell OEM extension to detach network iso.

python idrac_ctl.py oem-detach

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional
from base import Singleton, ApiRequestType, IDracManager, CommandResult


class DellOemDetach(IDracManager,
                    scm_type=ApiRequestType.DellOemDetach,
                    name='delloem_detach',
                    metaclass=Singleton):
    """A command uses dell oem to attach ISO
    """

    def __init__(self, *args, **kwargs):
        super(DellOemDetach, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser(is_remote_share=False)
        help_text = "command detach network iso"
        return cmd_parser, "oem-detach", help_text

    def execute(self,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Executes dell oem detach
        :param do_async: note async will subscribe to an event loop.
        :param verbose: enables verbose output
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        cmd_result = self.sync_invoke(ApiRequestType.DellOemActions, "dell_oem_actions")
        redfish_action = cmd_result.discovered['DetachISOImage']
        target_api = redfish_action.target

        api_result = self.base_post(target_api,
                                    do_async=do_async, expected_status=200)
        return CommandResult(api_result.data, None, None)



# TODO
# /redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/Actions/DellRaidService.ChangePDState
# /redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/Actions/DellRaidService.GetAvailableDisks
# /redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/Actions/DellRaidService.GetRAIDLevels
# /redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/Actions/DellRaidService.PrepareToRemove
# /redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/Actions/DellRaidService.RebuildPhysicalDisk



"""iDRAC reset a chassis power state.

python idrac_ctl.py chassis-reset --reset_type ForceOff

Author Mus spyroot@gmail.com
"""
import asyncio
import json
from abc import abstractmethod
from typing import Optional

from base import IDracManager, ApiRequestType, Singleton
from base import CommandResult, InvalidArgument
from base import FailedDiscoverAction, PostRequestFailed


class ChassisReset(IDracManager,
                   scm_type=ApiRequestType.ChassisReset,
                   name='reboot',
                   metaclass=Singleton):
    """
    """

    def __init__(self, *args, **kwargs):
        super(ChassisReset, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser(is_file_save=False, is_expanded=False)
        cmd_parser.add_argument('--reset_type',
                                required=False, dest='reset_type',
                                default="On", type=str,
                                help="On, ForceOff. On will change power state on.")

        help_text = "command reset chassis"
        return cmd_parser, "chassis-reset", help_text

    def execute(self,
                reset_type: Optional[str] = "On",
                do_async: Optional[bool] = False,
                data_type: Optional[str] = "json",
                **kwargs
                ) -> CommandResult:
        """
        :param do_async:
        :param data_type:
        :param reset_type: "On, ForceOff"
        :param kwargs:
        :return: return cmd result
        :raise FailedDiscoverAction
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        chassis_data = self.sync_invoke(ApiRequestType.ChassisQuery, "chassis_service_query")
        redfish_action = chassis_data.discovered['Reset']

        target_api = redfish_action.target
        args = redfish_action.args
        args_options = args['ResetType']
        if reset_type not in args_options:
            raise InvalidArgument(f"Unsupported reset type {reset_type} "
                                  f"supported reset options {args_options}.")

        if target_api is None:
            FailedDiscoverAction("Failed discover reset chassis actions.")

        payload = {'ResetType': reset_type}
        r = f"https://{self.idrac_ip}{target_api}"

        ok = False
        try:
            if not do_async:
                response = self.api_post_call(r, json.dumps(payload), headers)
                ok = self.default_post_success(self, response)
            else:
                loop = asyncio.get_event_loop()
                ok, response = loop.run_until_complete(self.async_post_until_complete(r, json.dumps(payload), headers))
        except PostRequestFailed as prf:
            print(prf)
            pass

        return CommandResult(self.api_success_msg(ok), None, None)



"""iDRAC query bios registry

Command provides raw query bios registry.
python idrac_ctl.py bios-registry --attr_name SystemServiceTag,OldSetupPassword

Will return SystemServiceTag,OldSetupPassword and list of all attributes.

python idrac_ctl.py bios-registry --attr_list --attr_name SystemServiceTag,OldSetupPassword

Author Mus spyroot@gmail.com
"""
import warnings
from abc import abstractmethod
from typing import Optional

import requests

from base import Singleton, ApiRequestType, IDracManager, CommandResult, UnexpectedResponse, InvalidArgument
from base.shared import ScheduleJobType
from datetime import datetime


class BiosChangeSettings(IDracManager, scm_type=ApiRequestType.BiosChangeSettings,
                         name='bios_change_settings',
                         metaclass=Singleton):
    """A command changes bios settings
    """

    def __init__(self, *args, **kwargs):
        super(BiosChangeSettings, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser(is_reboot=True, is_expanded=False)
        cmd_parser.add_argument('--attr_name', type=str,
                                required=True, dest="attr_name", metavar="attribute name",
                                default=None,
                                help="attribute name or list. Example --attr_name MemTest,EmbSata")

        cmd_parser.add_argument('--attr_value', type=str,
                                required=True, dest="attr_value", metavar="attribute value",
                                default=None,
                                help="attribute name or list. Example --attr_values Disabled,RaidMode")

        cmd_parser.add_argument('apply', choices=['on-reset', 'auto-boot', 'maintenance'],
                                default="on-reset",
                                help="choose when to apply change.")

        cmd_parser.add_argument("--start_date",
                                help="The start of maintenance window format YYYY-MM-DD",
                                required=False,
                                dest="start_date", metavar="start date",
                                type=str)

        cmd_parser.add_argument("--start_time",
                                help="The start of maintenance window format HH:MM:SS",
                                required=False, default="00:00:00",
                                dest="start_time", metavar="start time",
                                type=str)

        cmd_parser.add_argument('--duration', type=int, required=False,
                                dest="default_duration",
                                metavar="duration",
                                default=600,
                                help="maximum duration for a maintenance window from a start time")

        help_text = "command change bios values"
        return cmd_parser, "bios-change", help_text

    @staticmethod
    def crete_bios_config(current_config, attr_name, attr_val) -> dict:
        """Create new config for a bios.
        :param current_config:
        :param attr_name:
        :param attr_val:
        :return:
        """

        bios_payload = {
            "Attributes": {
            }
        }

        attribute_names = []
        attribute_values = []

        if attr_name is not None and len(attr_name) > 0:
            if "," in attr_name:
                attribute_names = attr_name.split(",")
            else:
                attribute_names = [attr_name]

        if attr_val is not None and len(attr_val) > 0:
            if "," in attr_val:
                attribute_values = attr_val.split(",")
            else:
                attribute_values = [attr_val]

        if len(attribute_names) != len(attribute_values):
            raise InvalidArgument("Number of attribute "
                                  "and values mismatched.")

        for name, val in zip(attribute_names, attribute_values):
            bios_payload["Attributes"][name.strip()] = val.strip()

        for k, v in bios_payload["Attributes"].items():
            for current_data in current_config:
                if k in current_data.values() and current_data['Type'] == "Integer":
                    bios_payload['Attributes'][k] = int(v)

        return bios_payload

    def execute(self,
                attr_name: Optional[str] = None,
                attr_value: Optional[str] = None,
                apply: Optional[str] = "on-reset",
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_reboot: Optional[bool] = False,
                start_date: Optional[str] = "",
                start_time: Optional[str] = "",
                default_duration: Optional[int] = 600,
                **kwargs) -> CommandResult:
        """Executes command to change bios settings.

        :param do_reboot:
        :param default_duration:
        :param apply:
        :param attr_value:  attribute value or list of values
        :param attr_name: attribute name or list of values
        :param do_async: note async will subscribe to an event loop.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param verbose: enables verbose output
        :param data_type: json or xml
        :param start_time:
        :param start_date:
        :return: CommandResult and if filename provide will save to a file.
        """
        target_api = "/redfish/v1/Systems/System.Embedded.1/Bios/BiosRegistry"
        cmd_result = self.base_query(target_api,
                                     filename=filename,
                                     do_async=do_async,
                                     do_expanded=False)

        registry = cmd_result.data['RegistryEntries']
        attribute_data = registry['Attributes']

        payload = self.crete_bios_config(attribute_data, attr_name, attr_value)
        if apply.strip() == "auto-boot":
            start_timestamp = datetime.fromisoformat(f'{start_date}T{start_time}')
            base_payload = self.schedule_job(ScheduleJobType.AutoReboot,
                                             start_time=str(start_timestamp),
                                             duration_time=default_duration)
        elif apply.strip() == "maintenance":
            start_timestamp = datetime.fromisoformat(f'{start_date}T{start_time}')
            base_payload = self.schedule_job(ScheduleJobType.NoReboot,
                                             start_time=str(start_timestamp),
                                             duration_time=default_duration)
        else:
            base_payload = self.schedule_job(ScheduleJobType.OnReset,
                                             start_time=None, duration_time=None)

        base_payload.update(payload)
        if verbose:
            print(f"payload: {base_payload}")

        target_api = "/redfish/v1/Systems/System.Embedded.1/Bios/Settings"
        api_result = self.base_patch(target_api, payload=payload,
                                     do_async=do_async, expected_status=200)
        if verbose:
            resp = api_result.extra
            print(f"api_result.data: hdr {resp.headers}")
            print(f"api_result.data: states {resp.status_code}")
            print(f"api_result.data: data {resp.json()}")

        result_data = api_result.data

        if api_result.extra is not None:
            resp = api_result.extra
            try:
                if isinstance(resp, requests.models.Response):
                    json_data = resp.json()
                    if verbose:
                        self.default_json_printer(json_data)

                    job_id = self.job_id_from_header(resp)
                    if job_id is not None:
                        data = self.fetch_job(job_id)
                        if isinstance(api_result.data, dict):
                            result_data.update(data)
            except UnexpectedResponse as ur:
                warnings.warn(str(ur))
                pass

        if do_reboot:
            cmd_result = self.sync_invoke(ApiRequestType.ChassisQuery,
                                          "chassis_service_query",
                                          data_filter="PowerState")

            if isinstance(cmd_result.data, dict) and 'PowerState' in cmd_result.data:
                pd_state = cmd_result.data['PowerState']
                if pd_state == 'On':
                    cmd_result = self.sync_invoke(ApiRequestType.RebootHost,
                                                  "reboot",
                                                  reset_type="ForceRestart")
                    if 'Status' in cmd_result.data:
                        result_data.update({"Reboot": cmd_result.data['Status']})
                else:
                    warnings.warn(f"Can't reboot a host , chassis power state {pd_state}")
            else:
                warnings.warn(f"Failed fetch chassis power state")

        return CommandResult({}, None, None)



"""iDRAC command return virtual media.

python idrac_ctl.py get_virtual_media

Command provides the option to retrieve virtual media from iDRAC
and serialize back to caller as JSON, YAML, or XML.
In addition, it automatically registered to the command line ctl tool.
Similarly to the rest command caller can save to a file and
consume asynchronously or synchronously.

python idrac_ctl.py get_virtual_media

- Each command return a result and list of REST Actions.
- Each command loaded based __init__ hence anyone can extend and add custom command.

Author Mus spyroot@gmail.com
"""
import argparse
from abc import abstractmethod
from typing import Optional

from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton
from base.cmd_utils import save_if_needed


class VirtualMediaGet(IDracManager, scm_type=ApiRequestType.VirtualMediaGet,
                      name='virtual_disk_query',
                      metaclass=Singleton):
    """iDRACs REST API Virtual Disk Query Command, fetch virtual disk, caller can save
    result to a file or output stdout or pass downstream to jq etc. tools.
    """
    def __init__(self, *args, **kwargs):
        super(VirtualMediaGet, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)
        cmd_arg.add_argument('-f', '--filename', required=False, type=str,
                             default="",
                             help="filename if we need to save a respond to a file.")

        help_text = "fetch the virtual media"
        return cmd_arg, "get_vm", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Execute command for virtual media query cmd.
        :param verbose: enables verbose output
        :param do_async: will not block and return result as future.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:  json, xml etc.
        :return: named tuple CommandResult
        :raise: AuthenticationFailed, UnexpectedResponse
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        if self.version_api():
            r = f"https://{self.idrac_ip}/redfish/v1/Systems/" \
                f"System.Embedded.1/VirtualMedia?$expand=*($levels=1)"
        else:
            r = f"https://{self.idrac_ip}/v1/Managers/iDRAC.Embedded.1/" \
                f"VirtualMedia?$expand=*($levels=1)"

        response = self.api_get_call(r, headers)
        self.default_error_handler(response)
        data = response.json()
        save_if_needed(filename, data)
        return CommandResult(data, None, None)



"""iDRAC query chassis services

Command provides raw query chassis and provide
list of supported actions.

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional
from base import Singleton, ApiRequestType, IDracManager, CommandResult


class VolumeQuery(IDracManager, scm_type=ApiRequestType.VolumeQuery,
                  name='vol_query',
                  metaclass=Singleton):
    """A command query job_service_query.
    """

    def __init__(self, *args, **kwargs):
        super(VolumeQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser()
        cmd_parser.add_argument('--dev_id',
                                required=True, dest="dev_id", type=str,
                                default=False,
                                help="storage controller Example. AHCI.Integrated.1-1")

        help_text = "command query volume from storage device."
        return cmd_parser, "volume-get", help_text

    def execute(self,
                dev_id: str,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Executes volume get action.

        :param dev_id:
        :param do_async: note async will subscribe to an event loop.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param verbose: enables verbose output
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        target_api = f"/redfish/v1/Systems/System.Embedded.1/Storage/{dev_id}/Volumes"
        cmd_result = self.base_query(target_api,
                                     filename=filename,
                                     do_async=do_async,
                                     do_expanded=True)

        actions = self.discover_member_redfish_actions(self, cmd_result.data)
        return CommandResult(cmd_result.data, actions, None)



"""iDRAC enable boot options.

This cmd return Dell Boot Sources Configuration and the related
resources.

Command provides the option to retrieve boot source from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command
caller can save to a file and consume asynchronously or synchronously.


Author Mus spyroot@gmail.com
"""
import json
import warnings

from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, CommandResult, UnexpectedResponse
from base.cmd_exceptions import InvalidArgument


class BootOneShot(IDracManager,
                  scm_type=ApiRequestType.BootOneShot,
                  name='boot_one_shot',
                  metaclass=Singleton):
    """
    Command enable boot option
    """

    def __init__(self, *args, **kwargs):
        super(BootOneShot, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser(is_reboot=True, is_expanded=False)

        cmd_parser.add_argument('--device', required=False, type=str,
                                default="Cd",
                                help="boot device Pxe,Cd,Hdd,BiosSetup,UefiTarget,SDCard etc")

        cmd_parser.add_argument('--uefi_target', required=False, type=str,
                                default=None,
                                help="uefi_target")

        help_text = "command change one shoot boot"
        return cmd_parser, "boot-one-shot", help_text

    def execute(self,
                device: Optional[str] = None,
                uefi_target: Optional[str] = None,
                do_check: Optional[str] = None,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_reboot: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Query information for particular boot source device from idrac.
        Example python idrac_ctl.py get_boot_source --dev "HardDisk.List.1-1"

        VenHw(986D1755-B9D0-4F8D-A0DA-D1DB18672045)

        :param do_reboot:  will reboot host
        :param uefi_target:
        :param device:  get the list of supported device.
                        For example None, Pxe,Cd,Hdd,BiosSetup,UefiTarget,SDCard,UefiHttp
        :param do_check:
        :param do_async: note async will subscribe to an event loop.
        :param verbose:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"do_async:{do_async} filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        current_boot = self.sync_invoke(ApiRequestType.CurrentBoot, "current_boot_query")
        boot_device = current_boot.data['BootSourceOverrideTarget@Redfish.AllowableValues']
        if device not in boot_device:
            raise InvalidArgument(f"Invalid boot device {device}, supported device {boot_device}")

        if uefi_target is not None:
            current_boot = self.sync_invoke(ApiRequestType.BootOptions, "boot_sources_query")
            self.default_json_printer(self, current_boot.extra)
            uefi_devs = [d['UefiDevicePath'] for d in current_boot.extra['Members'] if 'UefiDevicePath' in d]
            if uefi_target not in uefi_devs:
                raise InvalidArgument(f"Invalid uefi device path {uefi_target},"
                                      f" supported uefi devices {boot_device}")

        payload = {"Boot": {
            "BootSourceOverrideTarget": device,
            "UefiTargetBootSourceOverride": uefi_target}}

        r = f"https://{self.idrac_ip}/redfish/v1/Systems/System.Embedded.1"
        for key, value in dict(payload['Boot']).items():
            if value is None:
                del payload['Boot'][key]

        response = self.api_patch_call(r, json.dumps(payload), headers)
        api_result = {}
        if self.default_patch_success(self, response):
            api_result = self.api_success_msg(True)

        try:
            json_data = response.json()
            if verbose:
                self.default_json_printer(json_data)
            job_id = self.job_id_from_header(response)
            if job_id is not None:
                data = self.fetch_job(job_id)
                api_result.update(data)
        except UnexpectedResponse as ur:
            warnings.warn(str(ur))
            pass

        if do_reboot:
            reboot_result = self.reboot()
            api_result.update(reboot_result)

        return CommandResult(api_result, None, None)



"""overwrites some argparse defaults methods.
mainly for formatting.

Author Mus spyroot@gmail.com
"""
import argparse
import textwrap as _textwrap


class CustomArgumentDefaultsHelpFormatter(argparse.HelpFormatter):
    """Help message formatter which adds default values to argument help.
    Only the name of this class is considered a public API. All the methods
    provided by the class are considered an implementation detail.
    """

    def _get_help_string(self, action):
        help_msg = action.help
        if '%(default)' not in action.help:
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help_msg += ' (default: %(default)s)'
        return help_msg

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
            # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)

    def _iter_indented_subactions(self, action):
        """Sort and ident command
        :param action:
        :return:
        """
        try:
            get_subactions = action._get_subactions
        except AttributeError:
            pass
        else:
            self._indent()
            if isinstance(action, argparse._SubParsersAction):
                for subaction in sorted(get_subactions(), key=lambda x: x.dest):
                    yield subaction
            else:
                for subaction in get_subactions():
                    yield subaction
            self._dedent()

    def _fill_text(self, text, width, indent):
        """
        :param text:
        :param width:
        :param indent:
        :return:
        """
        text = self._whitespace_matcher.sub(' ', text).strip()
        paragraphs = text.split('|n ')
        multiline_text = ''
        for paragraph in paragraphs:
            formatted_paragraph = _textwrap.fill(paragraph, width,
                                                 initial_indent=indent,
                                                 subsequent_indent=indent) + '\n'
            multiline_text = multiline_text + formatted_paragraph
        return multiline_text



"""iDRAC reset a power state for compute system command

This action is used to reset the system.
Command provides the option to reboot, and change power state.

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
import json
import warnings
from abc import abstractmethod
from typing import Optional

from base import IDracManager, ApiRequestType, Singleton, CommandResult, UnexpectedResponse
from base.cmd_exceptions import InvalidArgument


class RebootHost(IDracManager,
                 scm_type=ApiRequestType.RebootHost,
                 name='reboot',
                 metaclass=Singleton):
    """
    "Actions": {
        "#ComputerSystem.Reset": {
            "ResetType@Redfish.AllowableValues": [
                "On",
                "ForceOff",
                "ForceRestart",
                "GracefulRestart",
                "GracefulShutdown",
                "PushPowerButton",
                "Nmi",
                "PowerCycle"
            ],
            "target": "/redfish/v1/Systems/System.Embedded.1/Actions/ComputerSystem.Reset"
        }
    },
    """
    def __init__(self, *args, **kwargs):
        super(RebootHost, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('--reset_type',
                                required=False, dest='reset_type',
                                default="GracefulRestart", type=str,
                                help="Reset On, ForceOff, "
                                     "ForceRestart, GracefulRestart, "
                                     "GracefulShutdown, "
                                     "PushPowerButton, Nmi, PowerCycle.")

        cmd_parser.add_argument('-a', '--async', action='store_true',
                                required=False, dest="do_async",
                                default=False, help="will use async call.")

        help_text = "reboots the system"
        return cmd_parser, "reboot", help_text

    def execute(self,
                filename: Optional[str] = "",
                data_type: Optional[str] = "json",
                reset_type: Optional[str] = "On",
                power_state: Optional[str] = "On",
                boot_source_override: Optional[str] = "",
                boot_source_override_enabled: Optional[str] = "",
                boot_source_override_mode: Optional[str] = "",
                interface_type: Optional[str] = "",
                do_async: Optional[bool] = False,
                **kwargs
                ) -> CommandResult:
        """
        :param do_async:
        :param filename:
        :param data_type:
        :param reset_type: "On, ForceOff, ForceRestart, GracefulShutdown, PushPowerButton, Nmi"
        :param power_state: On, null
        :param boot_source_override: "None, Pxe, Floppy,
                                      Cd, Hdd, BiosSetup, Utilities,
                                      UefiTarget, SDCard, UefiHttp"
        :param boot_source_override_enabled: "Once, Continuous, Disabled"
        :param boot_source_override_mode: UEFI, Legacy
        :param interface_type: TCM1_0, TPM2_0, TPM1_
        :param kwargs:
        :return:
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        system_state = self.sync_invoke(ApiRequestType.SystemQuery, "system_query")
        system_actions = system_state.data['Actions']
        allowed_reset_types = []
        if '#ComputerSystem.Reset' in system_actions:
            ra = system_actions['#ComputerSystem.Reset']
            allowed_reset_types = ra['ResetType@Redfish.AllowableValues']

        if reset_type not in allowed_reset_types:
            raise InvalidArgument(f"Invalid reset type"
                                  f" {reset_type}, "
                                  f"supported reset types "
                                  f"{allowed_reset_types}")

        t = "/redfish/v1/Systems/System.Embedded.1Actions/ComputerSystem.Reset"
        r = f"https://{self.idrac_ip}{t}"

        payload = {'ResetType': reset_type}
        if not do_async:
            response = self.api_post_call(r, json.dumps(payload), headers)
            ok = self.default_post_success(self, response)
        else:
            loop = asyncio.get_event_loop()
            ok, response = loop.run_until_complete(
                self.async_post_until_complete(r, json.dumps(payload), headers)
            )

        try:
            job_id = self.job_id_from_header(response)
            if job_id is not None:
                self.fetch_job(job_id)
        except UnexpectedResponse as ur:
            warnings.warn(str(ur))
            pass

        return CommandResult(self.api_success_msg(ok), None, None)



"""iDRAC storage controller query

Command provides the option to retrieve storage controller information
from iDRAC and serialize back as caller as JSON, YAML, or XML. In addition,
it automatically registered to the command line ctl tool. Similarly to
the rest command caller can save to a file and consume asynchronously
or synchronously.

On return will hold list of controller in CommandResult.data

['NonRAID.Slot.6-1',
'AHCI.Embedded.1-1',
 'AHCI.Slot.4-1',
 'AHCI.Embedded.2-1']

python idrac_ctl.py storage

Example: filter by controller type
idrac_ctl.py --json storage --filter AHCI

[
    "AHCI.Embedded.1-1",
    "AHCI.Slot.4-1",
    "AHCI.Embedded.2-1"
]

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
from abc import abstractmethod
from typing import Optional, Tuple

from base import Singleton, ApiRequestType, IDracManager, save_if_needed, CommandResult


class StorageQuery(IDracManager, scm_type=ApiRequestType.StorageQuery,
                   name='storage_query',
                   metaclass=Singleton):
    """
    Command return storage controller
    """

    def __init__(self, *args, **kwargs):
        super(StorageQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls) -> Tuple[argparse.ArgumentParser, str, str]:
        """Register command and all command optional flags.
        :param cls:
        :return: tuple ArgumentParser, command name, help_text
        """
        cmd_parser = cls.base_parser()

        cmd_parser.add_argument('-s', '--save_all', required=False, type=str, dest="do_save",
                                default=False,
                                help="for deep walk by default we don't "
                                     "save result to a file. save_all "
                                     "will save to a separate file.")

        cmd_parser.add_argument('--filter', required=False, type=str, dest="id_filter",
                                default="",
                                help="Filter based on controller information. "
                                     "Example --filter AHCI , "
                                     "will filter and only AHCI controllers")

        help_text = "command fetch the storage information"
        return cmd_parser, "storage-controllers", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                do_expanded: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                id_filter: Optional[str] = "",
                **kwargs) -> CommandResult:
        """Query storage controller from idrac.

         Example of members
         "Members": [
                 {"@odata.id": "/redfish/v1/Systems/System.Embedded.1/Storage/NonRAID.Slot.6-1"},
                 {"@odata.id": "/redfish/v1/Systems/System.Embedded.1/Storage/AHCI.Embedded.1-1"},
                 {"@odata.id": "/redfish/v1/Systems/System.Embedded.1/Storage/AHCI.Slot.4-1"},
                 {"@odata.id": "/redfish/v1/Systems/System.Embedded.1/Storage/AHCI.Embedded.2-1"}
                 ]
        :param filename:
        :param data_type:
        :param do_deep:
        :param do_expanded:
        :param verbose:
        :param do_async:
        :param id_filter:
        :param kwargs:
        :return:
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"do_deep:{do_deep} do_async:{do_async} id_filter:{id_filter}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        target_api = "/redfish/v1/Systems/System.Embedded.1/Storage"
        if do_expanded:
            r = f"https://{self.idrac_ip}{target_api}{self.expanded()}"
        else:
            r = f"https://{self.idrac_ip}{target_api}"

        if not do_async:
            response = self.api_get_call(r, headers)
            self.default_error_handler(response)
        else:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

        data = response.json()
        save_if_needed(filename, data)

        if id_filter is not None and len(id_filter) > 0:
            controller_list = [id_data['@odata.id'].split("/")[-1] for id_data in data['Members']
                               if id_filter in id_data['@odata.id']]

            controller_uri = [id_data['@odata.id'] for id_data in data['Members']
                              if id_filter in id_data['@odata.id']]

        else:
            controller_list = [id_data['@odata.id'].split("/")[-1] for id_data in data['Members']]
            controller_uri = [id_data['@odata.id'] for id_data in data['Members']]

        return CommandResult(controller_list, controller_uri, None)



"""iDRAC query bios registry

Command provides raw query bios registry.
python idrac_ctl.py bios-registry --attr_name SystemServiceTag,OldSetupPassword

Will return SystemServiceTag,OldSetupPassword and list of all attributes.

python idrac_ctl.py bios-registry --attr_list --attr_name SystemServiceTag,OldSetupPassword

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional
from base import Singleton, ApiRequestType, IDracManager, CommandResult, save_if_needed


class BiosRegistry(IDracManager,
                   scm_type=ApiRequestType.BiosRegistry,
                   name='bios_registry',
                   metaclass=Singleton):
    """A command query job_service_query.
    """

    def __init__(self, *args, **kwargs):
        super(BiosRegistry, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser()
        cmd_parser.add_argument('--registry_only',
                                action='store_true',
                                required=False, dest="is_registry_only",
                                default=True,
                                help="return only registry")

        cmd_parser.add_argument('--attr_only',
                                action='store_true',
                                required=False, dest="is_attr_only",
                                default=False,
                                help="return only attribute")

        cmd_parser.add_argument('--attr_name', type=str,
                                required=False, dest="attr_name",
                                default=None,
                                help="attribute name or list. Example --attr_name SystemServiceTag,OldSetupPassword")

        cmd_parser.add_argument('--attr_list', action='store_true',
                                required=False, dest="attr_list",
                                default=False,
                                help="return list of all attribute names.")

        cmd_parser.add_argument('--filter-read_only', action='store_true',
                                required=False, dest="no_read_only",
                                default=False,
                                help="will filter out all read-only.")

        help_text = "command query bios registry"
        return cmd_parser, "bios-registry", help_text

    def execute(self,
                is_registry_only: Optional[bool] = False,
                is_attr_only: Optional[bool] = False,
                attr_name: Optional[str] = None,
                attr_list: Optional[bool] = False,
                no_read_only: Optional[bool] = False,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_expanded: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Executes query bios registry.

        :param no_read_only:  filters, remove all read only.
        :param attr_list: will also return list of all attributes
        :param attr_name: will filler based on attribute names.
        :param is_registry_only: will return complete registry. (Default)
        :param is_attr_only: will return all attributes
        :param do_async: note async will subscribe to an event loop.
        :param do_expanded:  will do expand query
        :param filename: if filename indicate call will save a bios setting to a file.
        :param verbose: enables verbose output
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        target_api = "/redfish/v1/Systems/System.Embedded.1/Bios/BiosRegistry"
        cmd_result = self.base_query(target_api,
                                     filename=None,
                                     do_async=do_async,
                                     do_expanded=do_expanded)

        filtered_result = []
        attribute_names = None

        registry = cmd_result.data['RegistryEntries']
        data = registry['Attributes']

        if attr_list:
            if isinstance(data, list):
                attribute_names = [d['AttributeName'] for d in data]

        if attr_name is not None and len(attr_name) > 0:
            if "," in attr_name:
                attr_names = attr_name.split(",")
            else:
                attr_names = [attr_name]

            # filter by attribute name
            if isinstance(data, list):
                for n in attr_names:
                    for entry in data:
                        if entry['AttributeName'] == n:
                            # entry_dict.update(entry)
                            filtered_result.append(entry)
                            break
                data = filtered_result
            elif isinstance(data, dict):
                data = data[attr_name]
        else:
            if is_registry_only:
                data = cmd_result.data['RegistryEntries']
            if is_registry_only:
                registry = cmd_result.data['RegistryEntries']
                data = registry['Attributes']

        filtered_read_only = []
        if no_read_only:
            if 'RegistryEntries' in data:
                source_data = registry['RegistryEntries']['Attributes']
            elif 'Attributes' in data:
                source_data = registry['Attributes']
            elif isinstance(data, list):
                source_data = data
            else:
                raise ValueError("unknown value.")

            for entry in source_data:
                if entry['ReadOnly']:
                    continue
                else:
                    filtered_read_only.append(entry)
            data = filtered_read_only

        save_if_needed(filename, data)

        return CommandResult(data, None, attribute_names)



"""iDRAC bios command

Command provides the option to retrieve BIOS setting from iDRAC  and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command caller can save
to a file and consume asynchronously or synchronously.

python idrac_ctl.py --json bios --filter SystemModelName
python idrac_ctl.py --json bios --filter --filter systemmodelname
python idrac_ctl.py --json bios --filter ProcCStates

Filter by multiple attributes, note you search done case in sensitive

python idrac_ctl.py --json bios --filter systemmodelname,ProcCStates
{
    "ProcCStates": "Disabled",
    "systemmodelname": "PowerEdge R740"
}

On time boot

"OneTimeBootMode": "Disabled",
"BootMode": "Bios",

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
from abc import abstractmethod
from typing import Optional
from base import IDracManager, Singleton
from base import ApiRequestType, CommandResult
from base.cmd_utils import save_if_needed, find_ids
from base.custom_argparser.customer_argdefault import CustomArgumentDefaultsHelpFormatter


class BiosQuery(IDracManager, scm_type=ApiRequestType.BiosQuery,
                name='bios_inventory',
                metaclass=Singleton):
    """Bios Query Command, fetch bios data, caller can save to a file
    or output to a file or pass downstream.
    """

    def __init__(self, *args, **kwargs):
        super(BiosQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls, parent=None):
        """Registers command args
        :return:
        """
        if parent is not None:
            cmd_arg = argparse.ArgumentParser(add_help=False,
                                              parents=[parent],
                                              description="query iDRAC bios information",
                                              formatter_class=CustomArgumentDefaultsHelpFormatter)
        else:
            cmd_arg = argparse.ArgumentParser(add_help=False,
                                              description="query iDRAC bios information",
                                              formatter_class=CustomArgumentDefaultsHelpFormatter)

        # prog = None,
        # usage = None,
        # description = None,

        cmd_arg.add_argument('--async', required=False, default=False,
                             action='store_true', dest="do_async",
                             help="Will use asyncio.")

        cmd_arg.add_argument('--attr_only', required=False, default=False,
                             action='store_true', dest='attr_only',
                             help="Will only show attributes.")

        cmd_arg.add_argument('--filter', required=False, type=str, dest="attr_filter",
                             metavar="BIOS_ATTRIBUTE",
                             help="will filter on bios attribute information. "
                                  "Example --filter ProcCStates , "
                                  "will filter and and show C-State.")

        cmd_arg.add_argument('--deep', default=False, required=False, action='store_true',
                             dest="do_deep", help="deep walk. will make a separate "
                                                  "rest api call for each discovered api.")

        cmd_arg.add_argument('-f', '--filename', required=False, default="",
                             type=str,
                             help="filename if we need to save a respond to a file.")

        help_text = "command fetch the bios information"
        return cmd_arg, "bios", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                attr_only: Optional[bool] = False,
                attr_filter: Optional[str] = "",
                **kwargs) -> CommandResult:
        """Query bios from iDRAC.

        :param attr_filter: filters by BIOS attributes.
        :param attr_only: Will only output attribute i.e. current bios settings.
        :param do_async: will use asyncio
        :param verbose: verbose output, mainly for debug.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param do_deep: deep walk
        :param data_type: default json
        :return:
        """
        idrac_api = "/redfish/v1/Systems/System.Embedded.1/Bios"

        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"do_deep:{do_deep} do_async:{do_async} "
                  f"attr_filter:{attr_filter}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r: str = f"https://{self.idrac_ip}{idrac_api}"
        if not do_async:
            response = self.api_get_call(r, headers)
            self.default_error_handler(response)
        else:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

        data = response.json()
        # list of action for bios
        action_dict = self.discover_redfish_actions(self, data)
        if attr_only is True and 'Attributes' in data:
            data = {'Attributes': data['Attributes']}

        # filter
        data = self.filter_attribute(self, data, attr_filter)

        # save data
        save_if_needed(filename, data)

        extra_data_dict = {}
        if do_deep:
            api_links = find_ids(data, "@odata.id")
            api_links = [u for u in api_links if idrac_api != u]
            for api_link in api_links:
                r = f"https://{self.idrac_ip}{api_link}"
                if not do_async:
                    response = self.api_get_call(r, headers)
                    self.default_error_handler(response)
                else:
                    loop = asyncio.get_event_loop()
                    response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

                extra_data_dict[api_link] = response.json()

        for d in extra_data_dict.values():
            act = self.discover_redfish_actions(self, d)
            action_dict.update(act)

        return CommandResult(data, action_dict, extra_data_dict)



"""iDRAC command clear bios pending values.

Command provides the option to clear all the BIOS
pending values.

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
import json
from abc import abstractmethod
from typing import Optional

from base import IDracManager, ApiRequestType, Singleton, CommandResult
from base.cmd_exceptions import FailedDiscoverAction
from base.idrac_manager import PostRequestFailed


class BiosClearPending(IDracManager, scm_type=ApiRequestType.BiosClearPending,
                       name='clear_pending',
                       metaclass=Singleton):
    """
    This cmd action is used to clear all BIOS pending
    values currently in iDRAC.
    """

    def __init__(self, *args, **kwargs):
        super(BiosClearPending, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)

        cmd_parser.add_argument('--async', default=False, required=False,
                                action='store_true', dest="do_async",
                                help="Will use asyncio.")

        help_text = "command clear bios pending values"
        return cmd_parser, "bios-clear-pending", help_text

    def execute(self,
                do_async: Optional[bool] = False,
                data_type: Optional[str] = "json",
                **kwargs
                ) -> CommandResult:
        """Execute clear BIOS pending values

        :param do_async:
        :param data_type:
        :param kwargs:
        :return:
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        cmd_result = self.sync_invoke(ApiRequestType.BiosQuery,
                                      "bios_inventory", do_deep=True)
        api_target = None
        if cmd_result.discovered is not None and 'ClearPending' in cmd_result.discovered:
            api_target = cmd_result.discovered['ClearPending'].target

        if api_target is None:
            raise FailedDiscoverAction("Failed discover clear pending bios action.")

        api_req_result = {}
        try:
            pd = {}
            r = f"https://{self.idrac_ip}{api_target}"
            if not do_async:
                response = self.api_post_call(r, json.dumps(pd), headers)
                ok = self.default_post_success(self, response, expected=200)
            else:
                loop = asyncio.get_event_loop()
                ok, response = loop.run_until_complete(self.async_post_until_complete(r, json.dumps(pd), headers))
            api_req_result = {"Status": ok}

        except PostRequestFailed as pf:
            print("Error:", pf)
            pass

        return CommandResult(api_req_result, None, None)



"""iDRAC manager command

Command provides the option to retrieve iDRAC manager view.
python idrac_ctl.py manager

Author Mus spyroot@gmail.com
"""
import argparse
from abc import abstractmethod
from typing import Optional
from base import CommandResult
from base import IDracManager, ApiRequestType, Singleton


class Manager(IDracManager, scm_type=ApiRequestType.ResetManager,
              name='manager',
              metaclass=Singleton):
    """iDRAC Attribute Query Command, fetch attribute data, caller can save to a file
    or output to a file or pass downstream.
    """

    def __init__(self, *args, **kwargs):
        super(Manager, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)

        cmd_arg.add_argument('--async', action='store_true', required=False, dest="do_async",
                             default=False, help="Will create a task and will not wait.")

        cmd_arg.add_argument('-f', '--filename', required=False, type=str,
                             default="",
                             help="filename if we need to save a respond to a file.")

        help_text = "fetch the attribute view"
        return cmd_arg, "manager", help_text

    def execute(self, filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Queries attributes from iDRAC.
        :param do_async:
        :param verbose:
        :param do_deep:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:
        :return:
        :raise: AuthenticationFailed, UnexpectedResponse
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r = f"https://{self.idrac_ip}/redfish/v1/Managers/iDRAC.Embedded.1"
        response = self.api_get_call(r, headers)
        data = response.json()
        redfish_actions = self.discover_redfish_actions(self, data)
        return CommandResult(data, redfish_actions, None)



"""iDRAC enable boot options.

This cmd return Dell Boot Sources Configuration and the related
resources.

Command provides the option to retrieve boot source from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command
caller can save to a file and consume asynchronously or synchronously.


Author Mus spyroot@gmail.com
"""
import argparse

from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, CommandResult


class GetCurrentBoot(IDracManager, scm_type=ApiRequestType.CurrentBoot,
                     name='current_boot_query',
                     metaclass=Singleton):
    """
    Command enable boot option
    """
    cmd_name = "current_boot_query"

    def __init__(self, *args, **kwargs):
        super(GetCurrentBoot, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('--async', action='store_true', required=False, dest="do_async",
                                default=False, help="Will create a task and will not wait.")

        cmd_parser.add_argument('-f', '--filename', required=False, type=str,
                                default="",
                                help="filename if we need to save a respond to a file.")

        help_text = "command fetch the boot source for device/devices"
        return cmd_parser, "current_boot", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Query information for boot settings.
        Example python idrac_ctl.py get_boot_source --dev "HardDisk.List.1-1"

        :param do_async: note async will subscribe to an event loop.
        :param verbose:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"do_async:{do_async} filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r = f"https://{self.idrac_ip}/redfish/v1/Systems/System.Embedded.1"
        response = self.api_get_call(r, headers)
        self.default_error_handler(response)
        data = response.json()
        if 'Boot' in data:
            data = data['Boot']

        return CommandResult(data, None, None)



"""iDRAC system command

Command provides the option to retrieve system view from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest of commands
caller can save to a file and consume asynchronously or synchronously.

Author Mus spyroot@gmail.com
"""
import argparse
from abc import abstractmethod
from typing import Optional
from base.idrac_manager import CommandResult, IDracManager
from base.shared import ApiRequestType, Singleton
from base.cmd_utils import save_if_needed


class SystemQuery(IDracManager, scm_type=ApiRequestType.SystemQuery,
                  name='system_query',
                  metaclass=Singleton):
    """This main compute system query rest call.

    By default, will output system view without going deeper.
    In case caller provide do_deep will execute each respected rest_api
    and aggregate result.
    """

    def __init__(self, *args, **kwargs):
        super(SystemQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register commands args
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('--sub_option', action='store_true',
                                required=False, dest='module',
                                help="fetch main compute system view.")
        # --deep sub-command.
        cmd_parser.add_argument('--deep', action='store_true', required=False, dest="do_deep",
                                default=False, help="deep walk. will make a separate "
                                                    "REST call for each rest api.")

        cmd_parser.add_argument('-f', '--filename', required=False, type=str,
                                default="",
                                help="filename if we need to save a respond to a file.")

        cmd_parser.add_argument('-s', '--save_all', required=False, type=str, dest="do_save",
                                default=False, help="for deep walk by default we don't "
                                                    "save result to a file. save_all "
                                                    "will save to a separate file.")

        cmd_parser.add_argument('--save_dir', required=False, type=str, dest="save_dir",
                                default=False, help="will save json files in separate directory.")

        cmd_parser.add_argument('--async', action='store_true', required=False, dest="do_async",
                                default=False, help="Will create a task and will not wait.")

        help_text = "command fetch the system view."
        return cmd_parser, "system", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                do_deep: Optional[bool] = False,
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                save_dir: Optional[str] = None, **kwargs) -> CommandResult:
        """Execute main system query. From here we see what system support.
        Power state / and overall view.

        Method return actual json and dict where each key is action
        and respected reset end point.

        Bios: /redfish/v1/Systems/System.Embedded.1/Bios
        Memory /redfish/v1/Systems/System.Embedded.1/Memory
        Processors /redfish/v1/Systems/System.Embedded.1/Processors
        SecureBoot /redfish/v1/Systems/System.Embedded.1/SecureBoot
        SimpleStorage /redfish/v1/Systems/System.Embedded.1/SimpleStorage
        Storage /redfish/v1/Systems/System.Embedded.1/Storage
        VirtualMedia /redfish/v1/Systems/System.Embedded.1/VirtualMedia
        EthernetInterfaces: /redfish/v1/Systems/System.Embedded.1/EthernetInterfaces
        NetworkInterfaces /redfish/v1/Systems/System.Embedded.1/NetworkInterfaces

        :param save_dir:
        :param do_async:
        :param verbose:
        :param do_deep: if caller indicate deep, method will perform deep call
                     for each end point.
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:  what data type we use to serialize data.
        :return: json or xml and dict that host type of action and rest api end point.
        """
        if verbose:
            print(f"filename {filename} data type: {data_type} "
                  f"do_deep: {do_deep}, do_async: {do_async}, "
                  f"save_dir: {save_dir}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r = f"https://{self.idrac_ip}/redfish/v1/Systems/System.Embedded.1"
        response = self.api_get_call(r, headers)
        self.default_error_handler(response)
        data = response.json()
        save_if_needed(filename, data, save_dir=save_dir)

        rest_endpoints = {}
        extra_data_dict = {}

        for k in data.keys():
            if isinstance(data[k], dict) and "@odata.id" in data[k]:
                sub_rest = data[k]["@odata.id"]
                rest_endpoints[k] = sub_rest
                # deep walk
                if do_deep:
                    r = f"https://{self.idrac_ip}{sub_rest}"
                    response = self.api_get_call(r, headers)
                    self.default_error_handler(response)
                    if verbose:
                        print(f"sending request {r} status code {response.status_code}")
                    extra_data_dict[k] = response.json()

        return CommandResult(data, rest_endpoints, extra_data_dict)



"""iDRAC raid service query

Command provides the option to retrieve raid service  from iDRAC
and serialize back as caller as JSON, YAML, and XML. In addition,
it automatically registers to the command line ctl tool. Similarly
to the rest command caller can save to a file and consume asynchronously
or synchronously.

In result command also store all discovered action.

For example.
{'AssignSpare': '/redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService/Actions/DellRaidService.AssignSpare'}

Author Mus spyroot@gmail.com
"""
import argparse
from abc import abstractmethod
from typing import Optional

from base import ApiRequestType, IDracManager, Singleton, CommandResult
from base.cmd_utils import save_if_needed


class RaidServiceQuery(IDracManager, scm_type=ApiRequestType.RaidServiceQuery,
                       name='raid_service_query',
                       metaclass=Singleton):
    """Raid service query Command, fetch raid  service data, caller can save to a file
    or output to a file or pass downstream.
    """

    def __init__(self, *args, **kwargs):
        super(RaidServiceQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)
        cmd_parser.add_argument('-f', '--filename', required=False, type=str,
                                default="",
                                help="filename if we need to save a respond to a file.")

        help_text = "fetch the bios information"
        return cmd_parser, "raid", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Command query raid service and returns raid service in data field
        of Command Result. The discovered named tuple store action
        and respected rest APIs.

        :param do_async:
        :param verbose:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type:
        :return: return raid service
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r = f"https://{self.idrac_ip}/redfish/v1/Dell/Systems/" \
            f"System.Embedded.1/DellRaidService"

        response = self.api_get_call(r, headers)
        self.default_error_handler(response)
        data = response.json()
        save_if_needed(filename, data)
        actions = data['Actions']
        action_dict = {}
        for a in actions:
            raid_action = a.split(".")
            if len(a) > 0 and 'target' in actions[a]:
                action_dict[raid_action[1]] = actions[a]['target']

        return CommandResult(data, action_dict, None)



"""Shared resource
"""
import json
from enum import auto, Enum
from typing import Optional
from json import JSONEncoder


class ApiRequestType(Enum):
    """Each commands enum.
    """
    ConvertToRaid = auto()
    ConvertNoneRaid = ()
    Drives = auto()
    VolumeInit = auto()
    VolumeQuery = auto()
    ImportOneTimeBoot = auto()
    DellOemTask = auto()
    DellOemDisconnect = auto()
    RemoteServicesRssAPIStatus = auto()
    DellLcQuery = auto()
    RemoteServicesAPIStatus = auto()
    JobRmDellServices = auto()
    JobDellServices = auto()
    TasksList = auto()
    JobWatch = auto()
    BiosChangeSettings = auto()
    BiosRegistry = auto()
    ChangeBootOrder = auto()
    GetNetworkIsoAttachStatus = auto()
    OemAttach = auto()
    DellOemActions = auto()
    QueryIdrac = auto()
    JobDel = auto()
    AttributeClearPending = auto()
    JobGet = auto()
    Jobs = auto()
    BootOptions = auto()
    SystemConfigQuery = auto()
    BiosQuery = auto()
    IDracQuery = auto()
    AttributesQuery = auto()
    BootQuery = auto()
    FirmwareQuery = auto()
    FirmwareInventoryQuery = auto()
    PciDeviceQuery = auto()
    RebootHost = auto()
    SystemQuery = auto()
    VirtualDiskQuery = auto()
    RaidServiceQuery = auto()
    EnableBootOptions = auto()
    StorageQuery = auto()
    Tasks = auto()
    GetTask = auto()
    ImportSystem = auto()
    ResetManager = auto()
    VirtualMediaGet = auto()
    VirtualMediaInsert = auto()
    VirtualMediaEject = auto()
    CurrentBoot = auto()
    BootOneShot = auto()
    StorageViewQuery = auto()
    StorageListQuery = auto()
    BiosClearPending = auto()
    BootOptionQuery = auto()
    BootSettingsQuery = auto()
    BootSourceClear = auto()
    JobServices = auto()
    ChassisQuery = auto()
    ChassisReset = auto()
    GetAttachStatus = auto()
    DellOemNetIsoBoot = ()
    DellOemDetach = auto()
    TaskGet = auto()


class ScheduleJobType(Enum):
    """Each commands enum.
    """
    NoReboot = auto()
    AutoReboot = auto()
    OnReset = auto()


class RedfishActionEncoder(JSONEncoder):
    """JSON decoder used to serialize nested dicts.
    """
    def default(self, obj):
        return obj.__dict__


class RedfishAction:
    """Action discovery encapsulate each action to RedfishAction.
    """
    def __init__(self,
                 action_name: Optional[str] = "",
                 target: Optional[str] = "",
                 full_redfish_name: Optional[str] = ""):
        """Action discovered from json respond.
        """
        super().__init__()
        self.action_name = action_name
        self.full_redfish_name = full_redfish_name
        self.target = target
        self.args = None

    def __iter__(self):
        yield from {
            "action_name": self.action_name,
            "full_redfish_name": self.full_redfish_name,
            "target": self.target,
            "args": self.args,
        }.items()

    def add_action_arg(self, arg_name, allowable_value):
        """Add action argument name and allowable values for
        arguments for each args.
        :param arg_name: redfish action argument name
        :param allowable_value: redfish action argument allowable values
        :return:
        """
        if self.args is None:
            self.args = {}
        self.args[arg_name] = allowable_value

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
                          sort_keys=True, indent=4)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return json.dumps(dict(self), ensure_ascii=False)

    def to_json(self):
        return json.dumps(dict(self), ensure_ascii=False)


class Singleton(type):
    """This base class for all action that singleton
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return:
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



"""iDRAC clear pending values.

Command provides the option to clear all the pending attributes values.

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
import json
from abc import abstractmethod
from typing import Optional

from base import IDracManager, ApiRequestType, Singleton, CommandResult, FailedDiscoverAction
from base.idrac_manager import PostRequestFailed


class AttributeClearPending(IDracManager,
                            scm_type=ApiRequestType.AttributeClearPending,
                            name='clear_pending',
                            metaclass=Singleton):
    """
    This cmd action is used to clear all the pending
    values currently in iDRAC.
    """
    def __init__(self, *args, **kwargs):
        super(AttributeClearPending, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False)

        cmd_parser.add_argument('--async', default=False, required=False,
                                action='store_true', dest="do_async",
                                help="Will use asyncio.")

        help_text = "command clear attribute pending values"
        return cmd_parser, "attr-clear-pending", help_text

    def execute(self,
                do_async: Optional[bool] = False,
                data_type: Optional[str] = "json",
                **kwargs
                ) -> CommandResult:
        """Execute clear pending command.
        :param do_async:
        :param data_type:
        :param kwargs:
        :return:
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        target = None
        attributes_cmd = self.sync_invoke(ApiRequestType.AttributesQuery,
                                          "attribute_inventory", do_deep=True)
        if isinstance(attributes_cmd.extra, list):
            for extra in attributes_cmd.extra:
                if 'Actions' in extra:
                    actions = extra['Actions']
                    if '#DellManager.ClearPending' in actions:
                        target = actions['#DellManager.ClearPending']['target']

        if target is None:
            raise FailedDiscoverAction("Failed discover clear pending attribute action.")

        api_req_result = {}
        try:
            pd = {}
            r = f"https://{self.idrac_ip}{target}"
            if not do_async:
                response = self.api_post_call(r, json.dumps(pd), headers)
                ok = self.default_post_success(self, response, expected=200)
            else:
                loop = asyncio.get_event_loop()
                ok = loop.run_until_complete(self.async_post_until_complete(r, json.dumps(pd), headers))

            api_req_result = {"Status": ok}
        except PostRequestFailed as pf:
            print("Error:", pf)
            pass

        return CommandResult(api_req_result, None, None)



"""iDRAC enable boot options.

Command provides the option to enable boot source,
so it is used during a boot process.

For example.
python idrac_ctl.py boot-source-enable --dev NIC.Slot.8-1 --enable yes

Example
idrac_ctl.py set_boot_source --dev NIC.Slot.8-1 --enable yes

Enables boot on nic slot 8-1

Author Mus spyroot@gmail.com
"""
import argparse
import asyncio
import json
from abc import abstractmethod
from typing import Optional

from base import Singleton, ApiRequestType, IDracManager, CommandResult
from base.cmd_utils import str2bool
from base.idrac_manager import PatchRequestFailed


class EnableBootOptions(IDracManager,
                        scm_type=ApiRequestType.EnableBootOptions,
                        name='boot_patch',
                        metaclass=Singleton):
    """
    Command enable boot option
    """

    def __init__(self, *args, **kwargs):
        super(EnableBootOptions, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = argparse.ArgumentParser(add_help=False,
                                             description="command enable as boot source a device.")

        cmd_parser.add_argument('--async', action='store_true', required=False, dest="do_async",
                                default=False,
                                help="Will create a task and will not wait.")

        cmd_parser.add_argument('-f', '--filename', required=False, type=str,
                                default="",
                                help="filename if we need to save a respond to a file.")

        cmd_parser.add_argument('--dev', required=True, dest="boot_source",
                                type=str, default=None, metavar="DEVICE",
                                help="fetch verbose information for a device. Example --dev NIC.Slot.8-1")

        cmd_parser.add_argument('--enable', required=True, metavar="yes",
                                dest="is_enabled", type=str2bool, nargs='?',
                                help="Enable or Disable target boot device. yes|no, true|false")

        help_text = "command enable the boot on a particular device."
        return cmd_parser, "boot-source-enable", help_text

    def execute(self,
                boot_source: str,
                is_enabled: bool,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Query boot source from idrac
        :param is_enabled:  Enable or disable target boot device.
        :param boot_source: A device  Example, HardDisk.List.1-1 , NIC.Slot.8-1 etc
        :param do_async:
        :param verbose:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"boot_source:{boot_source}  is_enabled:{is_enabled}, "
                  f"do_async:{do_async} filename:{filename}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        if boot_source is None:
            raise ValueError("Please indicate boot source.")

        dev_result = self.sync_invoke(ApiRequestType.BootOptions, "boot_sources_query")
        target_dev = None
        boot_data_sources = []
        for full_dev_path in dev_result.data:
            devs = full_dev_path.split("/")
            if len(devs) > 0:
                dev = devs[-1]
                boot_data_sources.append(dev)
                if boot_source.lower() in dev.lower():
                    target_dev = full_dev_path
                    break

        if target_dev is None:
            print(f"Unknown dev. available boot source. {boot_data_sources}")
            return CommandResult(None, None, None)

        ok = False
        json_data = {}
        try:
            r = f"https://{self.idrac_ip}{target_dev}"
            payload = {"BootOptionEnabled": bool(is_enabled)}
            if not do_async:
                resp = self.api_patch_call(r, json.dumps(payload), hdr=headers)
                ok = self.default_patch_success(self, resp)
            else:
                loop = asyncio.get_event_loop()
                ok, resp = loop.run_until_complete(
                        self.api_async_patch_until_complete(
                                r, json.dumps(payload), headers)
                )

                json_data = resp.json()
                if '@Message.ExtendedInfo' in json_data:
                    json_data = json_data['@Message.ExtendedInfo']

        except PatchRequestFailed as patch_err:
            print("Error:", patch_err)
            pass
        except Exception as err:
            print("Error:", err)
            pass

        api_result = {"Status": ok}
        api_result.update(json_data)
        return CommandResult(api_result, None, None)



"""iDRAC query command

Command provides raw query option based URI resource, in
case specific action might not implement yet; hence it
is easy to query.

Author Mus spyroot@gmail.com
"""
from abc import abstractmethod
from typing import Optional
from base import Singleton, ApiRequestType, IDracManager, CommandResult


class QueryIDRAC(IDracManager, scm_type=ApiRequestType.QueryIdrac,
                 name='query_idrac',
                 metaclass=Singleton):
    """A command query iDRAC resource based on a resource path.
    """

    def __init__(self, *args, **kwargs):
        super(QueryIDRAC, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Register command and all optional flags.
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser()
        cmd_parser.add_argument('-r', '--resource', required=True, dest="resource",
                                type=str, default=None,
                                help="Job id. Example /redfish/v1/Managers")
        help_text = "command query based on resource."
        return cmd_parser, "query", help_text

    def execute(self, resource: str,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                do_expanded: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Executes query command
        python idrac_ctl.py query

        :param resource: path to a resource
        :param do_async: note async will subscribe to an event loop.
        :param do_expanded:  will do expand query
        :param filename: if filename indicate call will save a bios setting to a file.
        :param verbose: enables verbose output
        :param data_type: json or xml
        :return: CommandResult and if filename provide will save to a file.
        """
        return self.base_query(resource,
                               filename=filename,
                               do_async=do_async,
                               do_expanded=do_expanded)



"""iDRAC pci device query command

Command provides the option to retrieve firmware setting from iDRAC and serialize
back as caller as JSON, YAML, and XML. In addition, it automatically
registers to the command line ctl tool. Similarly to the rest command caller can save
to a file and consume asynchronously or synchronously.

Author Mus spyroot@gmail.com
"""
import asyncio
from abc import abstractmethod
from typing import Optional

from tqdm import tqdm

from base import Singleton, ApiRequestType, IDracManager, save_if_needed, CommandResult


class PciDeviceQuery(IDracManager,
                     scm_type=ApiRequestType.PciDeviceQuery,
                     name='pci_device_query',
                     metaclass=Singleton):
    """
    Command implementation to get pci device and pci functions.
    """

    def __init__(self, *args, **kwargs):
        super(PciDeviceQuery, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """
        :param cls:
        :return:
        """
        cmd_parser = cls.base_parser()
        cmd_parser.add_argument('pci_type', choices=['PCIeDevices', 'PCIeFunctions'],
                                default="PCIeDevices",
                                help="either pci device or pci function")

        help_text = "command fetch the pci device or function"
        return cmd_parser, "pci", help_text

    def execute(self,
                filename: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                pci_type: Optional[str] = "PCIeDevices",
                **kwargs) -> CommandResult:
        """Query pci device or function from idrac.
        :param verbose:
        :param do_async:
        :param filename: if filename indicate call will save a bios setting to a file.
        :param data_type: a data serialized back.
        :param pci_type: PCIeDevices or  PCIeFunctions
        :return: in data type json will return json
        """

        if verbose:
            print(f"cmd args data_type: {data_type} "
                  f"filename:{filename} do_async:{do_async} dev:{pci_type}")
            print(f"the rest of args: {kwargs}")

        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        r = f"https://{self.idrac_ip}/redfish/v1/Systems/" \
            f"System.Embedded.1?$select={pci_type}"

        if not do_async:
            response = self.api_get_call(r, headers)
            self.default_error_handler(response)
        else:
            loop = asyncio.get_event_loop()
            response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))

        data = response.json()
        pci_data = []
        if pci_type in data:
            api_endpoints = data[pci_type]
            with tqdm(total=len(api_endpoints)) as pbar:
                for r in api_endpoints:
                    r = f"https://{self.idrac_ip}{r['@odata.id']}"
                    if not do_async:
                        response = self.api_get_call(r, headers)
                        self.default_error_handler(response)
                    else:
                        loop = asyncio.get_event_loop()
                        response = loop.run_until_complete(self.api_async_get_until_complete(r, headers))
                    pci_data.append(response.json())
                    pbar.update(1)

        save_if_needed(filename, pci_data)
        return CommandResult(pci_data, None, None)



"""iDRAC virtual media eject command

Command provides the option to insert virtual disk with ISO image.
For example late we can set boot from virtual medial,
set one shot boot via one_shot cmd and reboot a host.

Example.  Retrieve list of current boot
python idrac_ctl.py current_boot

Note BootSourceOverrideEnabled and BootSourceOverrideTarget empty.

{
    "BootOptions": {
        "@odata.id": "/redfish/v1/Systems/System.Embedded.1/BootOptions"
    },
    "BootOrder": [
        "HardDisk.List.1-1",
        "NIC.Integrated.1-1-1",
        "NIC.Slot.8-1",
        "NIC.Slot.8-1",
        "NIC.Slot.8-1",
        "NIC.Slot.8-1",
        "Optical.iDRACVirtual.1-1"
    ],
    "BootOrder@odata.count": 7,
    "BootSourceOverrideEnabled": "Disabled",
    "BootSourceOverrideMode": "Legacy",
    "BootSourceOverrideTarget": "None",
    "BootSourceOverrideTarget@Redfish.AllowableValues": [
        "None",
        "Pxe",
        "Floppy",
        "Cd",
        "Hdd",
        "BiosSetup",
        "Utilities",
        "UefiTarget",
        "SDCard",
        "UefiHttp"
    ],
    "Certificates": {
        "@odata.id": "/redfish/v1/Systems/System.Embedded.1/Boot/Certificates"
    },
    "UefiTargetBootSourceOverride": null
}

python idrac_ctl.py get_virtual_media

We note that Image is null and not inserted.

"ConnectedVia": "NotConnected",
            "Description": "iDRAC Virtual Media Instance",
            "Id": "1",
            "Image": null,
            "ImageName": null,
            "Inserted": false,
            "MediaTypes": [
                "CD",
                "DVD",
                "USBStick"
            ],
            "MediaTypes@odata.count": 3,
            "Name": "VirtualMedia Instance 1",
            "Password": null,
            "TransferMethod": null,
            "TransferProtocolType": null,
            "UserName": null,
            "WriteProtected": true

Insert virtual media
python idrac_ctl.py insert_virtual_media --uri_path http://my_ip/ubuntu-22.04.1-desktop-amd64.iso --device_id 1

Author Mus spyroot@gmail.com
"""
import argparse
import json
import warnings
from abc import abstractmethod
from typing import Optional

from base import CommandResult, InvalidArgument
from base import IDracManager, ApiRequestType, Singleton


class VirtualMediaInsert(IDracManager,
                         scm_type=ApiRequestType.VirtualMediaInsert,
                         name='virtual_disk_insert',
                         metaclass=Singleton):
    """iDRACs cmd insert virtual media
    Virtual medial must be empty, otherwise command will throw exception.
    Called must first eject existing virtual media.
    """

    def __init__(self, *args, **kwargs):
        super(VirtualMediaInsert, self).__init__(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def register_subcommand(cls):
        """Registers command args
        :param cls:
        :return:
        """
        cmd_arg = argparse.ArgumentParser(add_help=False)

        cmd_arg.add_argument('--remote_username', required=False, type=str,
                             default=None,
                             help="remote username for authentication if required")

        cmd_arg.add_argument('--remote_password', required=False, type=str,
                             default=None,
                             help="remote password for authentication if required")

        cmd_arg.add_argument('--uri_path', required=False, type=str,
                             default=None,
                             help="url path to iso file. Example http://1.1.1.1/test.iso")

        cmd_arg.add_argument('--device_id', required=False, type=str,
                             default="1",
                             help="Default device id. Example 1 or 2")

        help_text = "command insert virtual media"
        return cmd_arg, "insert_vm", help_text

    def execute(self,
                uri_path: Optional[str] = None,
                remote_username: Optional[str] = None,
                remote_password: Optional[str] = None,
                device_id: Optional[str] = None,
                data_type: Optional[str] = "json",
                verbose: Optional[bool] = False,
                do_async: Optional[bool] = False,
                **kwargs) -> CommandResult:
        """Execute command virtual media eject
        :param device_id: virtual media device id 1 or 1
        :param remote_username:  username for remote authentication
        :param remote_password:  password for remote authentication
        :param uri_path: URI path to image file.
        :param verbose: enables verbose output
        :param do_async: will not block and return result as future.
        :param data_type:  json, xml etc.
        :return: named tuple CommandResult
        :raise: AuthenticationFailed, UnexpectedResponse
        """
        headers = {}
        if data_type == "json":
            headers.update(self.json_content_type)

        new_api = False
        virtual_media = self.sync_invoke(ApiRequestType.VirtualMediaGet,
                                         "virtual_disk_query")
        if self.version_api():
            new_api = True
            # this for a future issue if old API doesn't work

        if new_api is False:
            warnings.warn("Detected old rest API.")

        members = virtual_media.data['Members']
        actions = [self.discover_redfish_actions(self, m) for m
                   in members if m['Id'] == device_id]

        if len(actions) == 0:
            dev_id = [m['Id'] for m in members]
            raise InvalidArgument(f"Invalid device id {device_id}, "
                                  f"support device id {dev_id}")

        # if another image already mounted.
        inserted = {'image': m['Image'] for
                    m in members if m['Id'] == device_id and m['Inserted']}

        if 'image' in inserted:
            raise InvalidArgument(f"Image {inserted['image']} "
                                  f"already inserted. Eject media first.")

        eject_rest = [a['InsertMedia'].target for a in actions][-1]
        r = f"https://{self.idrac_ip}{eject_rest}"

        payload = {
            'Image': uri_path,
            'Inserted': True,
            'WriteProtected': True,
            'UserName': remote_username,
            'Password': remote_password,
        }

        for key, value in dict(payload).items():
            if value is None:
                del payload[key]

        response = self.api_post_call(r, json.dumps(payload), headers)
        ok = self.default_post_success(self, response, expected=204)
        return CommandResult(self.api_success_msg(ok), None, None)
